"""Connectors for cycle."""

import time as _time
from math import isclose as _isclose

import numpy as _np
from epics import PV as _PV

from ..util import get_bit as _get_bit
from ..namesys import SiriusPVName as _PVName
from ..envars import VACA_PREFIX
from ..search import PSSearch as _PSSearch, LLTimeSearch as _LLTimeSearch
from ..pwrsupply.csdev import Const as _PSConst, ETypes as _PSet, \
    PS_LI_INTLK_THRS as _PS_LI_INTLK
from ..timesys.csdev import Const as _TIConst, \
    get_hl_trigger_database as _get_trig_db

from .util import pv_timed_get as _pv_timed_get, pv_conn_put as _pv_conn_put, \
    get_trigger_by_psname as _get_trigger_by_psname, Const as _Const, \
    TRIGGER_NAMES as _TRIGGER_NAMES
from .bo_cycle_data import DEFAULT_RAMP_NRCYCLES, DEFAULT_RAMP_TOTDURATION, \
    bo_get_default_waveform as _bo_get_default_waveform
from .li_cycle_data import li_get_default_waveform as _li_get_default_waveform
from .fc_cycle_data import fc_get_default_waveform as _fc_get_default_waveform


TIMEOUT_SLEEP = 0.1
TIMEOUT_CONNECTION = 0.05
TIMEOUT_CHECK = 20


class Timing:
    """Timing."""

    # NOTE: this could be a class derived from one of the Device classes.

    EVTNAME_CYCLE = 'Cycle'

    DEFAULT_DURATION = 150  # [us]
    DEFAULT_NRPULSES = 1
    DEFAULT_DELAY = 0  # [us]
    DEFAULT_POLARITY = _TIConst.TrigPol.Normal

    _pvs = dict()
    properties = dict()
    evg_name = ''

    def __init__(self):
        """Init."""
        self._init_properties()
        self._initial_state = dict()
        self._create_pvs()

    @staticmethod
    def connected(sections=None, return_disconn=False):
        """Return connected state."""
        if sections is None:
            sections = list()
        disconn = set()
        for name, pvobj in Timing._pvs.items():
            name = _PVName(name)
            if (name.dev == 'EVG' or name.sec in sections) and\
                    not pvobj.connected:
                disconn.add(pvobj.pvname)
        if return_disconn:
            return disconn
        return not bool(disconn)

    # ----- main commands -----

    def prepare(self, mode, triggers=None):
        """Initialize properties."""
        if triggers is None:
            triggers = list()

        # Enable EVG
        self.enable_evg()

        # Disable Injection
        self.set_injection_state(_TIConst.DsblEnbl.Dsbl)

        # Config. triggers and events
        pvs_2_init = self.get_pvname_2_defval_dict(mode, triggers)
        for prop, defval in pvs_2_init.items():
            pvobj = Timing._pvs[prop]
            if prop.endswith(('Mon', )):
                continue
            if pvobj.wait_for_connection(TIMEOUT_CONNECTION):
                pvobj.value = defval
                _time.sleep(1.5*TIMEOUT_SLEEP)

        # Update events
        self.update_events()

    def check(self, mode, triggers=None):
        """Check if timing is configured."""
        if triggers is None:
            triggers = list()

        pvs_2_init = self.get_pvname_2_defval_dict(mode, triggers)
        for prop, defval in pvs_2_init.items():
            prop_sts = prop.replace('SP', 'RB').replace('Sel', 'Sts')
            pvobj = Timing._pvs[prop_sts]

            if not pvobj.wait_for_connection(TIMEOUT_CONNECTION):
                return False

            if prop_sts.propty_name.endswith(('Duration', 'Delay')):
                tol = 0.008 * 15
                if not _isclose(pvobj.value, defval, abs_tol=tol):
                    # print(pvobj.pvname, pvobj.value, defval)
                    return False
            elif isinstance(defval, (_np.ndarray, list, tuple)):
                if _np.any(pvobj.value[0:len(defval)] != defval):
                    # print(pvobj.pvname, pvobj.value, defval)
                    return False
            elif pvobj.value != defval:
                # print(pvobj.pvname, pvobj.value, defval)
                return False
        return True

    def trigger(self, mode):
        """Trigger timming to cycle power supply."""
        if mode == 'Cycle':
            pvobj = Timing._pvs[Timing.evg_name+':CycleExtTrig-Cmd']
            pvobj.value = 1
        elif mode == 'Ramp':
            pvobj = Timing._pvs[Timing.evg_name+':InjectionEvt-Sel']
            pvobj.value = _TIConst.DsblEnbl.Enbl

            pvobj = Timing._pvs[Timing.evg_name+':InjectionEvt-Sts']
            time0 = _time.time()
            while _time.time() - time0 < TIMEOUT_CHECK*3:
                if pvobj.value == _TIConst.DsblEnbl.Enbl:
                    break
                _time.sleep(TIMEOUT_SLEEP)
        else:
            raise NotImplementedError(
                "Trigger control is not defined for '{}' mode!".format(mode))

    def enable_evg(self):
        """Enable EVG."""
        pvobj = Timing._pvs[Timing.evg_name+':DevEnbl-Sts']
        if pvobj.value == _TIConst.DsblEnbl.Enbl:
            return
        pv_sel = Timing._pvs[Timing.evg_name+':DevEnbl-Sel']
        pv_sel.value = _TIConst.DsblEnbl.Enbl

    def set_injection_state(self, state):
        """Turn on/off InjectionEvt-Sel."""
        pvobj = Timing._pvs[Timing.evg_name+':InjectionEvt-Sel']
        pvobj.value = state

    def set_triggers_state(self, state, triggers, timeout=TIMEOUT_CHECK):
        """Set triggers state."""
        if isinstance(state, str):
            state = _TIConst.DsblEnbl.Enbl if state == 'enbl' \
                else _TIConst.DsblEnbl.Dsbl

        for trig in triggers:
            pvobj = Timing._pvs[trig+':State-Sel']
            pvobj.value = state
            _time.sleep(1.5*TIMEOUT_SLEEP)

        triggers_2_check = set(triggers)
        _t0 = _time.time()
        while _time.time() - _t0 < timeout:
            for trig in triggers:
                if trig not in triggers_2_check:
                    continue
                if Timing._pvs[trig+':State-Sts'].value == state:
                    triggers_2_check.remove(trig)
            if not triggers_2_check:
                break
            _time.sleep(TIMEOUT_SLEEP)
        if triggers_2_check:
            return False
        return True

    def update_events(self):
        """Update events."""
        pvobj = Timing._pvs[Timing.evg_name+':UpdateEvt-Cmd']
        pvobj.value = 1

    def get_cycle_count(self):
        """Get InjCount value."""
        pvobj = Timing._pvs[Timing.evg_name+':InjCount-Mon']
        return pvobj.value

    def check_ramp_end(self):
        """Check InjCount == DEFAULT_RAMP_NRCYCLES."""
        pvobj = Timing._pvs[Timing.evg_name+':InjCount-Mon']
        return pvobj.value == DEFAULT_RAMP_NRCYCLES

    def save_initial_state(self):
        """Save initial state."""
        for pvname in Timing._pvs:
            if pvname.propty_suffix in ('Cmd', 'Mon', 'Sts', 'RB'):
                continue
            self._initial_state[pvname] = Timing._pvs[pvname].value

    def restore_initial_state(self):
        """Restore initial state."""
        # Config. triggers and events to initial state
        for pvname, init_val in self._initial_state.items():
            if init_val is None:
                continue
            if ':InjectionEvt-Sel' in pvname:
                inj_state = init_val
                continue
            if ':BucketList-SP' in pvname and isinstance(init_val, int):
                init_val = [init_val, ]
            Timing._pvs[pvname].put(init_val)
            _time.sleep(1.5*TIMEOUT_SLEEP)

        # Update events
        self.update_events()

        # Set initial injection state
        self.set_injection_state(inj_state)

    def turnoff(self, triggers):
        """Turn timing off."""
        if triggers is None:
            triggers = list()

        # disable triggers
        pv_event = Timing._pvs[Timing.evg_name+':CycleMode-Sel']
        pv_event.value = _TIConst.EvtModes.Disable
        pv_bktlist = Timing._pvs[Timing.evg_name+':RepeatBucketList-SP']
        pv_bktlist.value = 0
        for trig in triggers:
            pvobj = Timing._pvs[trig+':Src-Sel']
            pvobj.value = 0  # Dsbl has always index 0
            pvobj = Timing._pvs[trig+':State-Sel']
            pvobj.value = _TIConst.DsblEnbl.Dsbl

    def get_pvnames_by_psnames(self, psnames=None):
        """Get pvnames to control according to psnames."""
        if psnames is None:
            psnames = list()
        triggers = _get_trigger_by_psname(psnames)
        pvnames = set()
        for mode in Timing.properties:
            for prop in Timing.properties[mode]:
                prop = _PVName(prop)
                if prop.dev == 'EVG' or prop.device_name in triggers:
                    pvnames.add(prop)
        return pvnames

    def get_pvname_2_defval_dict(self, mode, triggers=None):
        """Get pvnames to default values dict."""
        if triggers is None:
            triggers = list()
        pvname_2_defval_dict = dict()
        for prop, defval in Timing.properties[mode].items():
            if defval is None:
                continue
            prop = _PVName(prop)
            if prop.dev == 'EVG' or prop.device_name in triggers:
                pvname_2_defval_dict[prop] = defval
        return pvname_2_defval_dict

    # ----- private methods -----

    def _create_pvs(self):
        """Create PVs."""
        Timing._pvs = dict()
        for dict_ in Timing.properties.values():
            for pvname in dict_:
                if pvname in Timing._pvs:
                    continue
                pvname = _PVName(pvname)
                Timing._pvs[pvname] = _PV(
                    pvname.substitute(prefix=VACA_PREFIX),
                    connection_timeout=TIMEOUT_CONNECTION)

                if pvname.propty_suffix in ('Cmd', 'Mon'):
                    continue

                self._initial_state[pvname] = Timing._pvs[pvname].value

                if pvname.propty_suffix == 'SP':
                    pvname_sts = pvname.substitute(propty_suffix='RB')
                elif pvname.propty_suffix == 'Sel':
                    pvname_sts = pvname.substitute(propty_suffix='Sts')
                else:
                    continue
                Timing._pvs[pvname_sts] = _PV(
                    pvname_sts.substitute(prefix=VACA_PREFIX),
                    connection_timeout=TIMEOUT_CONNECTION)

    @classmethod
    def _init_properties(cls):
        if cls.properties:
            return
        cls.evg_name = _LLTimeSearch.get_evg_name()
        props = {
            'Cycle': {
                # EVG settings
                cls.evg_name+':DevEnbl-Sel': None,
                cls.evg_name+':DevEnbl-Sts': None,
                cls.evg_name+':InjectionEvt-Sel': None,
                cls.evg_name+':UpdateEvt-Cmd': None,

                # Cycle event settings
                cls.evg_name+':CycleMode-Sel': _TIConst.EvtModes.External,
                cls.evg_name+':CycleDelayType-Sel': _TIConst.EvtDlyTyp.Fixed,
                cls.evg_name+':CycleDelay-SP': cls.DEFAULT_DELAY,
                cls.evg_name+':CycleExtTrig-Cmd': None,
            },
            'Ramp': {
                # EVG settings
                cls.evg_name+':DevEnbl-Sel': None,
                cls.evg_name+':DevEnbl-Sts': None,
                cls.evg_name+':InjectionEvt-Sel': None,
                cls.evg_name+':BucketList-SP': [1, ],
                cls.evg_name+':RepeatBucketList-SP': DEFAULT_RAMP_NRCYCLES,
                cls.evg_name+':InjCount-Mon': None,
                cls.evg_name+':UpdateEvt-Cmd': None,

                # Cycle event settings
                cls.evg_name+':CycleMode-Sel': _TIConst.EvtModes.Injection,
                cls.evg_name+':CycleDelayType-Sel': _TIConst.EvtDlyTyp.Incr,
                cls.evg_name+':CycleDelay-SP': cls.DEFAULT_DELAY,
            }
        }

        for trig in _TRIGGER_NAMES:
            _trig_db = _get_trig_db(trig)
            _evt_index = _trig_db['Src-Sel']['enums'].index(cls.EVTNAME_CYCLE)
            for mode in ('Cycle', 'Ramp'):
                props[mode][trig+':Src-Sel'] = _evt_index
                props[mode][trig+':Duration-SP'] = cls.DEFAULT_DURATION
                props[mode][trig+':NrPulses-SP'] = cls.DEFAULT_NRPULSES
                props[mode][trig+':Delay-SP'] = cls.DEFAULT_DELAY
                props[mode][trig+':Polarity-Sel'] = cls.DEFAULT_POLARITY
                props[mode][trig+':State-Sel'] = None
                props[mode][trig+':Status-Mon'] = 0

        cls.properties = props


class PSCycler:
    """Handle power supplies properties related to Cycle and RmpWfm modes."""

    # NOTE: this could be a class derived from one of the Device classes.

    properties = [
        'Current-SP', 'Current-RB', 'CurrentRef-Mon',
        'PwrState-Sel', 'PwrState-Sts',
        'OpMode-Sel', 'OpMode-Sts',
        'CycleType-Sel', 'CycleType-Sts',
        'CycleFreq-SP', 'CycleFreq-RB',
        'CycleAmpl-SP', 'CycleAmpl-RB',
        'CycleOffset-SP', 'CycleOffset-RB',
        'CycleNrCycles-SP', 'CycleNrCycles-RB',
        'CycleAuxParam-SP', 'CycleAuxParam-RB',
        'CycleEnbl-Mon',
        'Wfm-SP', 'Wfm-RB',
        'WfmIndex-Mon', 'WfmSyncPulseCount-Mon',
        'IntlkSoft-Mon', 'IntlkHard-Mon',
        'SyncPulse-Cmd'
    ]

    def __init__(self, psname, ramp_config=None):
        """Constructor."""
        self._psname = _PVName(psname)
        self._ramp_config = ramp_config
        self._waveform = None
        self._siggen = None
        self._init_wfm_pulsecnt = None
        self._pvs = dict()
        for prop in self.properties:
            if prop not in self._pvs.keys():
                self._pvs[prop] = _PV(
                    self._psname.substitute(prefix=VACA_PREFIX, propty=prop),
                    connection_timeout=TIMEOUT_CONNECTION)

    @property
    def psname(self):
        """Power supply name."""
        return self._psname

    @property
    def connected(self):
        """Return connection state."""
        for prop in PSCycler.properties:
            if not self[prop].connected:
                return False
        return True

    def wait_for_connection(self, timeout=0.5):
        """Wait for connection."""
        for pvobj in self._pvs.values():
            if not pvobj.wait_for_connection(timeout):
                return False
        return True

    @property
    def waveform(self):
        """Return default waveform."""
        if self._waveform is None:
            self._waveform = _bo_get_default_waveform(
                psname=self.psname, ramp_config=self._ramp_config)
        return self._waveform

    @property
    def siggen(self):
        """Return default siggen."""
        if self._siggen is None:
            self._siggen = _PSSearch.conv_psname_2_siggenconf(self._psname)
        return self._siggen

    @property
    def init_wfm_pulsecnt(self):
        """Return initial waveform sync pulse count."""
        if self._init_wfm_pulsecnt is None:
            self.update_wfm_pulsecnt()
        return self._init_wfm_pulsecnt

    def update_wfm_pulsecnt(self):
        """Update waveform sync pulse count to current value."""
        self._init_wfm_pulsecnt = self._pvs['WfmSyncPulseCount-Mon'].get()

    def cycle_duration(self, mode):
        """Return the duration of the cycling in seconds."""
        if mode == 'Cycle':
            return self.siggen.num_cycles/self.siggen.freq
        if mode == 'Ramp':
            return DEFAULT_RAMP_TOTDURATION
        raise NotImplementedError(
            "Cycle duration is not defined for '{}' mode!".format(mode))

    def check_intlks(self, wait=2):
        """Check Interlocks."""
        wait = wait/2
        status = _pv_timed_get(self['IntlkSoft-Mon'], 0, wait=wait)
        status &= _pv_timed_get(self['IntlkHard-Mon'], 0, wait=wait)
        return status

    def check_on(self):
        """Return whether power supply PS is on."""
        return _pv_timed_get(self['PwrState-Sts'], _PSConst.PwrStateSts.On)

    def set_current_zero(self):
        """Set PS current to zero ."""
        return _pv_conn_put(self['Current-SP'], 0)

    def check_current_zero(self, wait=5):
        """Return whether power supply PS current is zero."""
        return _pv_timed_get(
            self['CurrentRef-Mon'], 0, abs_tol=0.05, wait=wait)

    def set_params(self, mode):
        """Set params to cycle."""
        if mode == 'Cycle':
            status = _pv_conn_put(self['CycleType-Sel'], self.siggen.sigtype)
            _time.sleep(TIMEOUT_SLEEP)
            status &= _pv_conn_put(self['CycleFreq-SP'], self.siggen.freq)
            _time.sleep(TIMEOUT_SLEEP)
            status &= _pv_conn_put(self['CycleAmpl-SP'], self.siggen.amplitude)
            _time.sleep(TIMEOUT_SLEEP)
            status &= _pv_conn_put(self['CycleOffset-SP'], self.siggen.offset)
            _time.sleep(TIMEOUT_SLEEP)
            status &= _pv_conn_put(self['CycleAuxParam-SP'],
                                   self.siggen.aux_param)
            _time.sleep(TIMEOUT_SLEEP)
            status &= _pv_conn_put(self['CycleNrCycles-SP'],
                                   self.siggen.num_cycles)
            _time.sleep(TIMEOUT_SLEEP)
        elif mode == 'Ramp':
            status = _pv_conn_put(self['Wfm-SP'], self.waveform)
            _time.sleep(TIMEOUT_SLEEP)
        else:
            raise NotImplementedError(
                "Parameters are not defined for mode '{}'!".format(mode))
        return status

    def check_params(self, mode, wait=5):
        """Return whether power supply cycling parameters are set."""
        if mode == 'Cycle':
            wait = wait/6
            type_idx = _PSet.CYCLE_TYPES.index(self.siggen.sigtype)
            status = _pv_timed_get(
                self['CycleType-Sts'], type_idx, wait=wait)
            status &= _pv_timed_get(
                self['CycleFreq-RB'], self.siggen.freq, wait=wait)
            status &= _pv_timed_get(
                self['CycleAmpl-RB'], self.siggen.amplitude, wait=wait)
            status &= _pv_timed_get(
                self['CycleOffset-RB'], self.siggen.offset, wait=wait)
            status &= _pv_timed_get(
                self['CycleAuxParam-RB'], self.siggen.aux_param, wait=wait)
            status &= _pv_timed_get(
                self['CycleNrCycles-RB'], self.siggen.num_cycles, wait=wait)
        elif mode == 'Ramp':
            status = _pv_timed_get(self['Wfm-RB'], self.waveform, wait=wait)
        else:
            raise NotImplementedError(
                "Parameters are not defined for mode '{}'!".format(mode))
        return status

    def prepare(self, mode):
        """Config power supply to cycling mode."""
        if not self.check_opmode_slowref(wait=1):
            return False

        if not self.check_current_zero(wait=1):
            return False

        status = self.set_params(mode)

        self.update_wfm_pulsecnt()
        return status

    def is_prepared(self, mode, wait=5):
        """Return whether power supply is ready."""
        if not self.check_current_zero(wait):
            return False
        if not self.check_params(mode, wait):
            return False
        return True

    def set_opmode(self, opmode):
        """Set power supply opmode to mode."""
        return _pv_conn_put(self['OpMode-Sel'], opmode)

    def set_opmode_slowref(self):
        """Set OpMode to SlowRef, if needed."""
        if self.check_opmode_slowref(wait=1):
            return True
        status = self.set_opmode(_PSConst.OpMode.SlowRef)
        _time.sleep(TIMEOUT_SLEEP)
        return status

    def check_opmode_slowref(self, wait=10):
        """Check if OpMode is SlowRef."""
        return _pv_timed_get(
            self['OpMode-Sts'], _PSConst.States.SlowRef, wait=wait)

    def set_opmode_cycle(self, mode):
        """Set OpMode to Cycle or RmpWfm according to mode."""
        opmode = _PSConst.OpMode.Cycle if mode == 'Cycle'\
            else _PSConst.OpMode.RmpWfm
        return self.set_opmode(opmode)

    def check_opmode_cycle(self, mode, wait=5):
        """Return whether power supply is in mode."""
        opmode = _PSConst.States.Cycle if mode == 'Cycle'\
            else _PSConst.States.RmpWfm
        return _pv_timed_get(self['OpMode-Sts'], opmode, wait=wait)

    def get_cycle_enable(self):
        """Check if Cycle is running."""
        if not self.connected:
            return False
        return self['CycleEnbl-Mon'].value == _PSConst.DsblEnbl.Enbl

    def pulse(self):
        """Send SyncPulse."""
        return _pv_conn_put(self['SyncPulse-Cmd'], 1)

    def check_final_state(self, mode):
        """Check state after Cycle."""
        if mode == 'Ramp':
            indices = len(self.waveform)
            status = _pv_timed_get(self['WfmIndex-Mon'], indices, wait=10.0)
            status &= _pv_timed_get(
                self['WfmSyncPulseCount-Mon'],
                self.init_wfm_pulsecnt + DEFAULT_RAMP_NRCYCLES, wait=10.0)
            self.update_wfm_pulsecnt()
            if not status:
                return _Const.CycleEndStatus.LackTriggers
        else:
            status = _pv_timed_get(self['CycleEnbl-Mon'], 0, wait=10.0)
            if not status:
                return _Const.CycleEndStatus.NotFinished

        status = self.check_intlks(wait=1.0)
        if not status:
            return _Const.CycleEndStatus.Interlock

        return _Const.CycleEndStatus.Ok

    def __getitem__(self, prop):
        """Return item."""
        return self._pvs[prop]


class PSCyclerFBP(PSCycler):
    """Handle PS FBP properties related to Cycle and RmpWfm modes."""

    properties = [
        'Current-SP', 'Current-RB', 'CurrentRef-Mon',
        'PwrState-Sel', 'PwrState-Sts',
        'OpMode-Sel', 'OpMode-Sts',
        'SOFBMode-Sel', 'SOFBMode-Sts',
        'CycleType-Sel', 'CycleType-Sts',
        'CycleFreq-SP', 'CycleFreq-RB',
        'CycleAmpl-SP', 'CycleAmpl-RB',
        'CycleOffset-SP', 'CycleOffset-RB',
        'CycleNrCycles-SP', 'CycleNrCycles-RB',
        'CycleAuxParam-SP', 'CycleAuxParam-RB',
        'CycleEnbl-Mon',
        'Wfm-SP', 'Wfm-RB',
        'WfmIndex-Mon', 'WfmSyncPulseCount-Mon',
        'IntlkSoft-Mon', 'IntlkHard-Mon',
        'SyncPulse-Cmd'
    ]

    def set_sofbmode(self, state):
        """Set SOFBMode."""
        state = _PSConst.OffOn.On if state == 'on' else _PSConst.OffOn.Off
        return _pv_conn_put(self['SOFBMode-Sel'], state)

    def check_sofbmode(self, state, wait=1):
        """Check if SOFBMode."""
        state = _PSConst.OffOn.On if state == 'on' else _PSConst.OffOn.Off
        return _pv_timed_get(self['SOFBMode-Sts'], state, wait=wait)

    def prepare(self, mode):
        """Config power supply to cycling mode."""
        if not self.check_sofbmode('off', wait=1):
            return False

        if not self.check_opmode_slowref(wait=1):
            return False

        if not self.check_current_zero(wait=1):
            return False

        status = self.set_params(mode)

        self.update_wfm_pulsecnt()
        return status


class LinacPSCycler:
    """Handle Linac power supply properties to cycle."""

    # NOTE: this could be a class derived from one of the Device classes.

    properties = [
        'Current-SP', 'Current-Mon', 'PwrState-Sel',
        'StatusIntlk-Mon', 'IntlkWarn-Mon',
    ]

    def __init__(self, psname):
        """Constructor."""
        self._psname = _PVName(psname)
        self._waveform = None
        self._cycle_duration = None
        self._times = None
        self._pvs = dict()
        self.intlkwarn_bit = _PSet.LINAC_INTLCK_WARN.index('LoadI Over Thrs')
        for prop in LinacPSCycler.properties:
            if prop not in self._pvs.keys():
                self._pvs[prop] = _PV(
                    self._psname.substitute(prefix=VACA_PREFIX, propty=prop),
                    connection_timeout=TIMEOUT_CONNECTION)

    @property
    def psname(self):
        """Power supply name."""
        return self._psname

    @property
    def connected(self):
        """Return connected state."""
        for prop in LinacPSCycler.properties:
            if not self[prop].connected:
                return False
        return True

    def wait_for_connection(self, timeout=0.5):
        """Wait for connection."""
        for pvobj in self._pvs.values():
            if not pvobj.wait_for_connection(timeout):
                return False
        return True

    @property
    def waveform(self):
        """Return waveform."""
        if self._waveform is None:
            self._get_duration_and_waveform()
        return self._waveform

    def cycle_duration(self, _):
        """Return the duration of the cycling in seconds."""
        if self._cycle_duration is None:
            self._get_duration_and_waveform()
        return self._cycle_duration

    def check_intlks(self, wait=2):
        """Check interlocks."""
        if not self.connected:
            return False
        intlkval = self['StatusIntlk-Mon'].value
        if self.psname.dev == 'Spect':
            intlkwarn = self['IntlkWarn-Mon'].value
            if _get_bit(intlkwarn, self.intlkwarn_bit):
                intlkval -= 2**self.intlkwarn_bit
        return intlkval < _PS_LI_INTLK

    def check_on(self):
        """Return whether power supply PS is on."""
        return _pv_timed_get(self['PwrState-Sel'], 1)

    def set_current_zero(self):
        """Set PS current to zero ."""
        return _pv_conn_put(self['Current-SP'], 0)

    def check_current_zero(self, wait=5):
        """Return whether power supply PS current is zero."""
        return _pv_timed_get(self['Current-Mon'], 0, abs_tol=0.1, wait=wait)

    def prepare(self, _):
        """Config power supply to cycling mode."""
        status = True
        if not self.check_current_zero(wait=0.5):
            status &= self.set_current_zero()
        return status

    def is_prepared(self, mode, wait=5):
        """Return whether power supply is ready."""
        return self.check_current_zero(wait)

    def cycle(self):
        """Cycle. This function may run in a thread."""
        for i in range(len(self._waveform)-1):
            self['Current-SP'].value = self._waveform[i]
            _time.sleep(self._times[i+1] - self._times[i])
        self['Current-SP'].value = self._waveform[-1]

    def check_final_state(self, _):
        """Check state after Cycle."""
        status = self.check_on()
        status &= self.check_intlks()
        if not status:
            return _Const.CycleEndStatus.Interlock
        return _Const.CycleEndStatus.Ok

    def _get_duration_and_waveform(self):
        """Get duration and waveform."""
        time, wfm = _li_get_default_waveform(psname=self.psname)
        self._times = time
        self._cycle_duration = max(time)
        self._waveform = wfm

    def __getitem__(self, prop):
        """Return item."""
        return self._pvs[prop]


class FOFBPSCycler:
    """Handle FOFB power supply properties to cycle."""

    # NOTE: this could be a class derived from one of the Device classes.

    properties = [
        'Current-SP', 'Current-Mon', 'PwrState-Sts',
        'AlarmsAmp-Mon', 'OpMode-Sel', 'OpMode-Sts',
    ]

    def __init__(self, psname):
        """Constructor."""
        self._psname = _PVName(psname)
        self._waveform = None
        self._cycle_duration = None
        self._times = None
        self._pvs = dict()
        for prop in FOFBPSCycler.properties:
            if prop not in self._pvs.keys():
                self._pvs[prop] = _PV(
                    self._psname.substitute(prefix=VACA_PREFIX, propty=prop),
                    connection_timeout=TIMEOUT_CONNECTION)

    @property
    def psname(self):
        """Power supply name."""
        return self._psname

    @property
    def connected(self):
        """Return connected state."""
        for prop in FOFBPSCycler.properties:
            if not self[prop].connected:
                return False
        return True

    def wait_for_connection(self, timeout=0.5):
        """Wait for connection."""
        for pvobj in self._pvs.values():
            if not pvobj.wait_for_connection(timeout):
                return False
        return True

    @property
    def waveform(self):
        """Return waveform."""
        if self._waveform is None:
            self._get_duration_and_waveform()
        return self._waveform

    def cycle_duration(self, _):
        """Return the duration of the cycling in seconds."""
        if self._cycle_duration is None:
            self._get_duration_and_waveform()
        return self._cycle_duration

    def check_intlks(self, wait=2):
        """Check interlocks."""
        if not self.connected:
            return False
        return _pv_timed_get(self['AlarmsAmp-Mon'], 0, wait=wait)

    def check_on(self):
        """Return whether power supply PS is on."""
        return _pv_timed_get(self['PwrState-Sts'], _PSConst.OffOn.On)

    def set_current_zero(self):
        """Set PS current to zero ."""
        return _pv_conn_put(self['Current-SP'], 0)

    def check_current_zero(self, wait=5):
        """Return whether power supply PS current is zero."""
        return _pv_timed_get(self['Current-Mon'], 0, abs_tol=0.01, wait=wait)

    def prepare(self, _):
        """Config power supply to cycling mode."""
        status = True
        if not self.check_current_zero(wait=0.5):
            status &= self.set_current_zero()
        return status

    def is_prepared(self, _, wait=5):
        """Return whether power supply is ready."""
        return self.check_current_zero(wait)

    def set_opmode_slowref(self):
        """Set OpMode to closed_loop_manual, if needed."""
        if self.check_opmode_slowref(wait=1):
            return True
        sts = _pv_conn_put(
            self['OpMode-Sel'],
            _PSConst.OpModeFOFB.closed_loop_manual)
        _time.sleep(TIMEOUT_SLEEP)
        return sts

    def check_opmode_slowref(self, wait=10):
        """Check if OpMode is closed_loop_manual."""
        return _pv_timed_get(
            self['OpMode-Sts'],
            _PSConst.OpModeFOFB.closed_loop_manual,
            wait=wait)

    def cycle(self):
        """Cycle. This function may run in a thread."""
        for i in range(len(self._waveform)-1):
            self['Current-SP'].value = self._waveform[i]
            _time.sleep(self._times[i+1] - self._times[i])
        self['Current-SP'].value = self._waveform[-1]

    def check_final_state(self, _):
        """Check state after Cycle."""
        status = self.check_on()
        status &= self.check_intlks()
        if not status:
            return _Const.CycleEndStatus.Interlock
        return _Const.CycleEndStatus.Ok

    def _get_duration_and_waveform(self):
        """Get duration and waveform."""
        time, wfm = _fc_get_default_waveform(psname=self.psname)
        self._times = time
        self._cycle_duration = max(time)
        self._waveform = wfm

    def __getitem__(self, prop):
        """Return item."""
        return self._pvs[prop]



"""This module contains classes for current normalization."""

import re as _re
import numpy as _np

import mathphys as _mp

from .. import util as _util
from ..namesys import SiriusPVName as _SiriusPVName
from ..search import PSSearch as _PSSearch
from ..search import MASearch as _MASearch
from ..pwrsupply.csdev import get_ps_propty_database as _get_ps_propty_database

from . import util as _mutil


# beta(energy) ~ 1 approximation is more computationally efficient.
_BETA_APPROXIMATION = True
if not _BETA_APPROXIMATION:
    _GAMMA_2_GEV = _mp.constants.electron_rest_energy * _mp.units.joule_2_GeV

_MAGFUNCS = _mutil.get_magfunc_2_multipole_dict()
_IS_DIPOLE = _re.compile(".*:[A-Z]{2}-B.*:.+$")
_IS_FAM = _re.compile(".*[A-Z]{2}-Fam:[A-Z]{2}-.+$")
_KCOEFF = _mp.constants.elementary_charge / \
          _mp.constants.light_speed / \
          _mp.constants.electron_mass / \
          2 / _np.pi  # [1/(T.m)]


class _MagnetNormalizer:
    """Base class for converting magnet properties: current and strength."""

    def __init__(self, maname, magnet_conv_sign=-1):
        """Class constructor."""
        self._maname = _SiriusPVName(maname) if type(maname) == str else maname
        self._psnames = _MASearch.conv_maname_2_psnames(self._maname)
        self._psmodel = _PSSearch.conv_psname_2_psmodel(self._psnames[0])
        self._pstype = _PSSearch.conv_psname_2_pstype(self._psnames[0])
        self._magfunc = _PSSearch.conv_pstype_2_magfunc(self._pstype)
        self._excdata = _PSSearch.conv_psname_2_excdata(self._psnames[0])
        self._magnet_conv_sign = magnet_conv_sign
        self._mfmult = _MAGFUNCS[self._magfunc]
        self._psname = self._power_supplies()[0]
        self._calc_conv_coef()

    @property
    def magfunc(self):
        """Return magnet function string."""
        return self._magfunc

    @property
    def maname(self):
        """Return magnet name."""
        return self._maname

    # --- normalizer interface ---

    def conv_current_2_strength(self, currents, **kwargs):
        """Convert current to strength."""
        if currents is None:
            return None
        intfields = self._conv_current_2_intfield(currents)
        # NOTE: really necessary? ---
        if intfields is None:
            if isinstance(currents, (int, float)):
                return 0.0
            else:
                return [0.0, ] * len(currents)
        # ---
        strengths = self._conv_intfield_2_strength(intfields, **kwargs)
        strengths = self._conv_default_2_epicsdb(strengths)
        return strengths

    def conv_strength_2_current(self, strengths, **kwargs):
        """Convert strength to current."""
        strengths = self._conv_epicsdb_2_default(strengths)
        intfields = self._conv_strength_2_intfield(strengths, **kwargs)
        mft = self._mfmult
        currents = self._excdata.interp_mult2curr(
            intfields, mft['harmonic'], mft['type'])
        return currents

    # --- normalizer aux. methods ---

    def _conv_current_2_intfield(self, currents):
        mpoles = self._conv_current_2_multipoles(
            currents, only_main_harmonic=True)
        if mpoles is None:
            return None
        mfm = self._mfmult
        intfield = mpoles[mfm['type']][mfm['harmonic']]
        return intfield

    def _conv_current_2_multipoles(self, currents, only_main_harmonic=False):
        # NOTE: think about implementation of this function for magnets with
        # multiple functions...
        if currents is None:
            return None
        msum = {}
        if self._magfunc != 'dipole':
            mpoles = \
                self._excdata.interp_curr2mult(currents, only_main_harmonic)
            msum = _mutil.sum_magnetic_multipoles(msum, mpoles)
        else:
            mpoles = \
                self._excdata.interp_curr2mult(currents, only_main_harmonic)
            msum = _mutil.sum_magnetic_multipoles(msum, mpoles)
        return msum

    def _power_supplies(self):
        psname = self._maname.replace(":MA", ":PS").replace(':PM', ':PU')
        return [psname]

    # --- conversion default [rad] to epicsdb [(m|u)rad] values ---

    def _conv_values(self, values, coef):
        if isinstance(values, (int, float)):
            return coef * values
        elif isinstance(values, tuple):
            return (coef*v for v in values)
        elif isinstance(values, list):
            return [coef*v for v in values]
        elif isinstance(values, _np.ndarray):
            return coef*values
        else:
            raise ValueError()

    def _conv_default_2_epicsdb(self, values):
        return self._conv_values(values, self._coef_def2edb)

    def _conv_epicsdb_2_default(self, values):
        return self._conv_values(values, 1.0/self._coef_def2edb)

    def _calc_conv_coef(self):
        dbase = _get_ps_propty_database(self._psmodel, self._pstype)
        if 'Energy-SP' in dbase:
            dbase = dbase['Energy-SP']
        elif 'KL-SP' in dbase:
            dbase = dbase['KL-SP']
        elif 'SL-SP' in dbase:
            dbase = dbase['SL-SP']
        elif 'Kick-SP' in dbase:
            dbase = dbase['Kick-SP']
        if 'unit' in dbase:
            unit = dbase['unit'].lower()
            if unit == 'mrad':
                self._coef_def2edb = 1e3
                return
            if unit == 'urad':
                self._coef_def2edb = 1e6
                return
            if unit == 'rad':
                self._coef_def2edb = 1
                return
            if unit == 'deg':
                self._coef_def2edb = 180.0 / _np.pi
                return
        self._coef_def2edb = 1.0

    def _conv_intfield_2_strength(self, intfields, **kwargs):
        raise NotImplementedError


class DipoleNormalizer(_MagnetNormalizer):
    """Convert magnet current to strength and vice versa."""

    TYPE = 'DipoleNormalizer'

    _ref_angles = _mutil.get_nominal_dipole_angles()

    def __init__(self, maname, **kwargs):
        """Class constructor."""
        super(DipoleNormalizer, self).__init__(maname, **kwargs)
        self._set_reference_dipole_data()

    # --- normalizer aux. methods ---

    def _set_reference_dipole_data(self):
        ang = DipoleNormalizer._ref_angles
        if self._maname.sec == 'SI':
            self._ref_energy = 3.0  # [GeV]
            self._ref_brho, self._ref_beta, self._ref_gamma, *_ = \
                _util.beam_rigidity(self._ref_energy)
            # self._ref_bl_bc = - self._ref_brho * ang['SI_BC']
            # self._ref_angle = ang['SI_B1'] + ang['SI_B2']  + ang['SI_BC']
            # self._ref_bl = - self._ref_brho * self._ref_angle \
            #     - self._ref_bl_bc
            self._ref_angle = ang['SI_B1'] + ang['SI_B2']
            self._ref_bl = - self._ref_brho * self._ref_angle
        elif self._maname.sec == 'BO':
            self._ref_energy = 3.0  # [GeV]
            self._ref_brho, self._ref_beta, self._ref_gamma, *_ = \
                _util.beam_rigidity(self._ref_energy)
            self._ref_angle = ang['BO']
            self._ref_bl = - self._ref_brho * self._ref_angle
        elif self._maname.sec == 'TS':
            self._ref_energy = 3.0  # [GeV]
            self._ref_brho, *_ = _util.beam_rigidity(self._ref_energy)
            self._ref_angle = ang['TS']
            self._ref_bl = - self._ref_brho * self._ref_angle
        elif self._maname.sec == 'TB':
            self._ref_energy = 0.150  # [GeV]
            self._ref_brho, *_ = _util.beam_rigidity(self._ref_energy)
            self._ref_angle = ang['TB']
            self._ref_bl = - self._ref_brho * self._ref_angle
        elif self._maname.sec == 'LI':
            self._ref_energy = 0.150  # [GeV]
            self._ref_brho, *_ = _util.beam_rigidity(self._ref_energy)
            self._ref_angle = ang['LI']
            self._ref_bl = - self._ref_brho * self._ref_angle
        else:
            raise NotImplementedError

    def _conv_strength_2_intfield(self, strengths, **kwargs):
        _ = kwargs  # throwaway arguments
        if isinstance(strengths, list):
            strengths = _np.array(strengths)
        if self._maname.sec == 'SI':
            if _BETA_APPROXIMATION:
                # 1. approximation beta(energy) ~ 1.0
                intfields = (- self._ref_angle *
                             (self._ref_brho / self._ref_energy) *
                             strengths)  # - self._ref_bl_bc)
            else:
                # 2. without approximation
                brho, *_ = _util.beam_rigidity(strengths)
                intfields = brho * (- self._ref_angle)  # - self._ref_bl_bc
        else:
            if _BETA_APPROXIMATION:
                # 1. approximation beta(energy) ~ 1.0
                intfields = (- self._ref_angle *
                             (self._ref_brho / self._ref_energy) * strengths)
            else:
                # 2. without approximation
                brho, *_ = _util.beam_rigidity(strengths)
                intfields = brho * (- self._ref_angle)

        return intfields

    def _conv_intfield_2_strength(self, intfields, **kwargs):
        if isinstance(intfields, list):
            intfields = _np.array(intfields)
        if self._maname.sec == 'SI':
            if _BETA_APPROXIMATION:
                # 1. approximation beta(energy) ~ 1.0
                total_bl = intfields  # + self._ref_bl_bc
                strengths = -self._magnet_conv_sign * \
                    ((self._ref_energy / self._ref_brho) *
                     (-total_bl) / self._ref_angle)
            else:
                # 2. without approximation
                total_bl = intfields  # + self._ref_bl_bc
                beam_rigidity = -total_bl / self._ref_angle
                alpha = (beam_rigidity / self._ref_brho) * \
                    (self._ref_gamma**2 - 1.0)/(self._ref_gamma)
                gamma = (alpha/2) + _np.sqrt(1.0 + (alpha/2)**2)
                strengths = gamma * _GAMMA_2_GEV
        else:
            if _BETA_APPROXIMATION:
                # 1. approximation beta(energy) ~ 1.0
                strengths = -self._magnet_conv_sign * \
                    ((self._ref_energy / self._ref_brho) *
                     (- intfields) / self._ref_angle)
            else:
                # 2. without approximation
                total_bl = intfields
                beam_rigidity = -total_bl / self._ref_angle
                alpha = (beam_rigidity / self._ref_brho) * \
                    (self._ref_gamma**2 - 1.0)/(self._ref_gamma)
                gamma = (alpha/2) + _np.sqrt(1.0 + (alpha/2)**2)
                strengths = gamma * _GAMMA_2_GEV

        return strengths

    def _power_supplies(self):
        return self._psnames


class MagnetNormalizer(_MagnetNormalizer):
    """Convert magnet current to strength and vice versa.

    Since we decided to match signs of Kick-Mon and direction
    of the beam kick, as we do in beam dynamic models, we have
    to treat horizontal and vertical correctors differently in the
    conversion from current to strength and vice-versa.
    """

    TYPE = 'MagnetNormalizer'

    def __init__(self, maname, **kwargs):
        """Call superclass init and initializes a dipole."""
        super(MagnetNormalizer, self).__init__(maname, **kwargs)

    def _conv_strength_2_intfield(self, strengths, **kwargs):
        if isinstance(strengths, list):
            strengths = _np.array(strengths)
        brho, *_ = _util.beam_rigidity(kwargs['strengths_dipole'])
        intfields = self._magnet_conv_sign * brho * strengths
        return intfields

    def _conv_intfield_2_strength(self, intfields, **kwargs):
        if isinstance(intfields, list):
            intfields = _np.array(intfields)
        brho, *_ = _util.beam_rigidity(kwargs['strengths_dipole'])
        if isinstance(brho, _np.ndarray):
            with _np.errstate(divide='ignore', invalid='ignore'):
                strengths = self._magnet_conv_sign * intfields / brho
                strengths[brho == 0] = 0.0
        else:
            if brho == 0:
                strengths = 0.0
            else:
                strengths = self._magnet_conv_sign * intfields / brho
        if not isinstance(intfields, (int, float)):
            if isinstance(strengths, (int, float)):
                strengths = [strengths, ] * len(intfields)
        return strengths


class APUNormalizer(_MagnetNormalizer):
    """."""

    def _conv_strength_2_intfield(self, strengths, **kwargs):
        """Convert K parameter to field amplitude.

        For APU, integrated field is just the field amplitude B * lamba [T.m].
        The strength is the K parameter:
            K ~ 93.3729/(T.m) * (lambda * B)
        """
        _ = kwargs  # throwaway arguments
        if isinstance(strengths, list):
            strengths = _np.array(strengths)

        intfields = strengths / _KCOEFF
        return intfields

    def _conv_intfield_2_strength(self, intfields, **kwargs):
        """Convert field amplitude to K parameter.

        For APU, integrated field is just the field amplitude B * lamba [T.m].
        The strength is the K parameter:
            K ~ 93.3729/(T.m) * (lambda * B)
        """
        _ = kwargs  # throwaway arguments
        if isinstance(intfields, list):
            intfields = _np.array(intfields)

        strengths = _KCOEFF * intfields
        return strengths


class TrimNormalizer(_MagnetNormalizer):
    """Convert trim magnet current to strength and vice versa."""

    TYPE = 'TrimNormalizer'

    def __init__(self, maname, magnet_conv_sign=-1.0,
                 **kwargs):
        """Call super and initializes a dipole and the family magnet."""
        super(TrimNormalizer, self).__init__(maname, **kwargs)

    def _conv_strength_2_intfield(self, strengths, **kwargs):
        if isinstance(strengths, list):
            strengths = _np.array(strengths)
        strengths_fam = kwargs['strengths_family']
        brho, *_ = _util.beam_rigidity(kwargs['strengths_dipole'])
        intfields = self._magnet_conv_sign * brho * (strengths - strengths_fam)
        return intfields

    def _conv_intfield_2_strength(self, intfields, **kwargs):
        if isinstance(intfields, (list, tuple)):
            intfields = _np.array(intfields)
        brho, *_ = _util.beam_rigidity(kwargs['strengths_dipole'])
        if brho == 0:
            return 0 * intfields
        strengths_trim = self._magnet_conv_sign * intfields / brho
        # integrated field in excitation data for trims is just
        # its contribution.
        strengths_fam = _np.array(kwargs['strengths_family'])
        return strengths_trim + strengths_fam



#!/usr/bin/env python-sirius

"""Test PRU module."""
from unittest import TestCase

from siriuspy import util
from siriuspy.pwrsupply.pructrl import pru
from siriuspy.pwrsupply.pructrl.pru import PRUInterface
from siriuspy.pwrsupply.pructrl.pru import PRU


PUB_INTERFACE = (
    'PRUInterface',
    'PRU',
)


class TestModule(TestCase):
    """Test module interface."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(pru, PUB_INTERFACE)
        self.assertTrue(valid)


class TestPRUInterface(TestCase):
    """Test PRUInterface API."""

    PUB_INTERFACE = (
        'open',
        'close',
        'UART_write',
        'UART_read',
        'UART_request',
        'wr_duration',
        'wr_duration_reset',
    )

    def test_public_interface(self):
        """Test class public interface."""
        self.assertTrue(util.check_public_interface_namespace(
            PRUInterface, TestPRUInterface.PUB_INTERFACE))

    def test_simulated(self):
        """Test simulated."""
        # TODO: implement test!
        pass

    def test_UART_write(self):
        """Test UART_write."""
        # TODO: implement test!
        pass

    def test_UART_read(self):
        """Test UART_read."""
        # TODO: implement test!
        pass

    def test_close(self):
        """Test close."""
        # TODO: implement test!
        pass


class TestPRU(TestCase):
    """Test PRU API."""

    PUB_INTERFACE = (
        'bbbname',
        'ip_address',
    )

    def test_public_interface(self):
        """Test class public interface."""
        self.assertTrue(util.check_public_interface_namespace(
            PRU, TestPRU.PUB_INTERFACE))

    def test_bbbname(self):
        """Test bbbname."""
        # TODO: implement test!
        pass

    def test_ip_address(self):
        """Test bbbname."""
        # TODO: implement test!
        pass

    def test_UART_write(self):
        """Test UART_write."""
        # TODO: implement test!
        pass

    def test_UART_read(self):
        """Test UART_read."""
        # TODO: implement test!
        pass

    def test_close(self):
        """Test close."""
        # TODO: implement test!
        pass




"""Utilities for cycle."""

import time as _time
import math as _math
import numpy as _np

from ..csdev import Const as _Const
from ..namesys import Filter as _Filter
from ..search import PSSearch as _PSSearch, HLTimeSearch as _HLTimeSearch


TRIGGER_NAMES = {
    'TB-Glob:TI-Mags', 'TS-Glob:TI-Mags',
    'BO-Glob:TI-Mags-Fams', 'BO-Glob:TI-Mags-Corrs',
    'SI-Glob:TI-Mags-Bends', 'SI-Glob:TI-Mags-Quads',
    'SI-Glob:TI-Mags-Sexts', 'SI-Glob:TI-Mags-Skews',
    'SI-Glob:TI-Mags-Corrs', 'SI-Glob:TI-Mags-QTrims'}


def get_psnames(isadv=False):
    """Return psnames."""
    names = _PSSearch.get_psnames({'sec': '(LI|TB|TS)', 'dis': 'PS'})

    if not isadv:
        names.extend(_PSSearch.get_psnames(
            {'sec': 'SI', 'sub': 'Fam', 'dis': 'PS', 'dev': '(B|Q.*|S.*)'}))
        names.extend(_PSSearch.get_psnames(
            {'sec': 'SI', 'sub': '[0-2][0-9]C2', 'dis': 'PS',
             'dev': 'CV', 'idx': '2'}))
        names.extend(_PSSearch.get_psnames(
            {'sec': 'SI', 'sub': '[0-2][0-9]C2', 'dis': 'PS',
             'dev': 'QS'}))
        names.extend(_PSSearch.get_psnames(
            {'sec': 'SI', 'dis': 'PS', 'dev': 'FC.*'}))
    else:
        names.extend(_PSSearch.get_psnames(
            {'sec': 'SI', 'dis': 'PS', 'dev': '(B|Q.*|S.*|C.*|FC.*)'}))

    to_remove = _PSSearch.get_psnames({'sec': 'TS', 'idx': '(0|1E2)'})
    for name in to_remove:
        names.remove(name)
    return names


def get_sections(psnames):
    """Return sections."""
    sections = list()
    for sec in ('LI', 'TB', 'BO', 'TS', 'SI'):
        if _Filter.process_filters(psnames, filters={'sec': sec}):
            sections.append(sec)
    return sections


def get_trigger_by_psname(psnames):
    """Return triggers corresponding to psnames."""
    psnames = set(psnames)
    triggers = set()
    for trig in TRIGGER_NAMES:
        dev_names = set(_HLTimeSearch.get_hl_trigger_channels(trig))
        dev_names = {dev.device_name for dev in dev_names}
        if psnames & dev_names:
            triggers.add(trig)
    return triggers


def pv_timed_get(pvobj, value, wait=5, abs_tol=0.0, rel_tol=1e-06):
    """Do timed get."""
    if not pvobj.connected:
        return False
    time0 = _time.time()
    while _time.time() - time0 < wait:
        pvvalue = pvobj.value
        if isinstance(value, (tuple, list, _np.ndarray)):
            if not isinstance(pvvalue, (tuple, list, _np.ndarray)):
                _time.sleep(wait/10.0)
                continue
            elif len(value) != len(pvvalue):
                _time.sleep(wait/10.0)
                continue
            else:
                if _np.allclose(pvvalue, value, atol=abs_tol, rtol=rel_tol):
                    return True
        else:
            if _math.isclose(pvvalue, value, abs_tol=abs_tol, rel_tol=rel_tol):
                return True
    return False


def pv_conn_put(pvobj, value):
    """Put if connected."""
    if not pvobj.connected:
        return False
    if pvobj.put(value):
        return True
    return False


class Const(_Const):
    """PSCycle Constants."""

    CycleEndStatus = _Const.register(
        'CycleEndStatus', ('Ok', 'LackTriggers', 'NotFinished', 'Interlock'))



"""Define the low level classes which will connect to Timing Devices IOC."""

import re as _re
from functools import partial as _partial
import logging as _log
from threading import Thread as _ThreadBase
from epics.ca import CASeverityException as _CASeverityException

from ..util import update_bit as _update_bit, get_bit as _get_bit
from ..epics import CONNECTION_TIMEOUT as _CONN_TIMEOUT, PV as _PV
from ..envars import VACA_PREFIX as LL_PREFIX
from ..namesys import SiriusPVName as _PVName
from ..search import LLTimeSearch as _LLSearch, HLTimeSearch as _HLSearch
from ..callbacks import Callback as _Callback
from ..devices import Event as _Event
from ..thread import RepeaterThread as _Timer

from .csdev import Const as _TIConst


_RFFREQ = _TIConst.RF_FREQUENCY
_RFDIV = _TIConst.RF_DIVISION
_ACFREQ = _TIConst.AC_FREQUENCY
_US2SEC = 1e-6
_FDEL = _TIConst.FINE_DELAY / _US2SEC


class _Thread(_ThreadBase):

    def __init__(self, **kwargs):
        if 'daemon' not in kwargs:
            kwargs['daemon'] = True
        super().__init__(**kwargs)


class _BaseLL(_Callback):

    def __init__(self, channel, prefix):
        """Initialize the Low Level object.

        callback is the callable to be called each time a low level PV changes
        its value.
        """
        super().__init__()
        self._initialized = False
        self.channel = _PVName(channel)
        self.prefix = prefix
        self._dict_functs_for_write = self._define_dict_for_write()
        self._dict_functs_for_update = self._define_dict_for_update()
        self._dict_functs_for_read = self._define_dict_for_read()
        self._dict_convert_prop2pv = self._define_convertion_prop2pv()
        self._dict_convert_pv2prop = {
            val: key for key, val in self._dict_convert_prop2pv.items()}
        self._config_ok_values = dict()
        self._base_freq = _RFFREQ / _RFDIV

        self._writepvs = dict()
        self._readpvs = dict()
        self._locked = False
        self._lock_threads_dict = dict()

        evts = _HLSearch.get_configurable_hl_events()
        self._events = {evt: _Event(evt) for evt in evts}

        evg_name = _PVName(_LLSearch.get_evg_name())
        self._base_freq_pv = _PV(
            evg_name.substitute(prefix=LL_PREFIX, propty='FPGAClk-Cte'))
        self._update_base_freq()
        self._base_freq_pv.add_callback(self._update_base_freq)

        for prop, pvname in self._dict_convert_prop2pv.items():
            pvnamerb = pvnamesp = None
            if not _PVName.is_write_pv(pvname):
                pvnamerb = pvname
                pvnamesp = _PVName.from_rb2sp(pvname)
            elif _PVName.is_cmd_pv(pvname):  # -Cmd is different!!
                self._writepvs[prop] = _PV(pvname)

            if pvnamerb is not None:
                pvo = _PV(pvnamerb)
                pvo.add_callback(self._on_change_readpv)
                pvo.connection_callbacks.append(self._on_connection)
                self._readpvs[prop] = pvo
            if pvnamesp != pvnamerb and not prop.endswith('DevEnbl'):
                pvo = _PV(pvnamesp)
                self._writepvs[prop] = pvo
                pvo.initialized = False
                pvo.add_callback(self._on_change_writepv)
                pvo.connection_callbacks.append(self._on_connection_writepv)
        self._initialized = True

    @property
    def connected(self):
        """."""
        pvs = list(self._readpvs.values()) + list(self._writepvs.values())
        pvs += [self._base_freq_pv, ]
        for evt in self._events.values():
            pvs += [evt.pv_object(p) for p in evt.properties]
        conn = True
        for pv in pvs:
            conn &= pv.connected
            if not pv.connected:
                _log.debug('NOT CONN: {0:s}'.format(pv.pvname))
        return conn

    def wait_for_connection(self, timeout=None):
        """."""
        pvs = list(self._readpvs.values()) + list(self._writepvs.values())
        pvs += [self._base_freq_pv, ]
        for evt in self._events.values():
            pvs += [evt.pv_object(p) for p in evt.properties]
        for pv in pvs:
            if not pv.wait_for_connection(timeout=timeout):
                _log.info(pv.pvname + ' not connected.')
                return False
        return True

    @property
    def locked(self):
        """."""
        return self._locked

    @locked.setter
    def locked(self, value):
        """."""
        self._set_locked(value)

    def write(self, prop, value):
        """Set property values in low level IOCS.

        Function called by classes that control high level PVs to transform
        the high level values into low level properties and set the low level
        IOCs accordingly.
        """
        fun = self._dict_functs_for_write.get(prop)
        if fun is None:
            _log.warning('No write function defined')
            return False
        dic = fun(value)  # dic must be None for -Cmd PVs
        if dic is None:
            return True
        elif isinstance(dic, dict) and not dic:
            _log.warning('Function return value is empty')
            return False  # dic is empty in case write was not successfull
        for prop, val in dic.items():
            self.write_ll(prop, val)
        return True

    def write_ll(self, prop, value):
        self._config_ok_values[prop] = value
        pv = self._writepvs.get(prop)
        if pv is None:
            return
        # I decided not to wait put to finish when writing on LL PVs
        # to make the high level IOC as fast as possible.
        # To guarantee that the desired value will be written on the
        # rather slow LL IOC, I put the setpoint in the verification
        # queue.
        self._put_on_pv(pv, value, wait=False)
        pvname = self._dict_convert_prop2pv[prop]
        self._start_lock_thread(pvname)

    def read(self, prop, is_sp=False):
        """Read HL properties from LL IOCs and return the value."""
        fun = self._dict_functs_for_read[prop]
        if fun is None:
            return None
        dic_ = fun(is_sp)
        if dic_ is None or prop not in dic_:
            return None
        return dic_[prop]

    def _set_locked(self, value):
        self._locked = bool(value)
        self.run_callbacks(self.channel, 'LowLvlLock', value, is_sp=False)
        if not self._locked:
            return
        for prop, val in self._config_ok_values.items():
            if val is None:
                continue
            self.write_ll(prop, val)

    def _update_base_freq(self, **kwargs):
        self._base_freq = self._base_freq_pv.get(
            timeout=_CONN_TIMEOUT) or self._base_freq
        self.base_del = 1 / self._base_freq / _US2SEC
        self._rf_del = self.base_del / 5

    def _define_convertion_prop2pv(self):
        """Define a dictionary for convertion of names.

        The dictionary converts low level properties names to low level PV
        names.
        """
        return dict()

    def _define_dict_for_write(self):
        """Define a dictionary of functions to convert HL to LL.

        Each function converts the values of High level properties into values
        of Low Level properties. The functions defined in this dictionary are
        called by write and they send to the Low Level IOC the converted
        values.
        """
        return dict()

    def _define_dict_for_update(self):
        """Define a dictionary of functions to convert LL to HL.

        Each function converts the readback values of Low Level properties into
        values of High Level properties and send them to the High Level classes
        for update of the readback PVs.
        It is called by the on_change_pvs_rb and calls the callback function.
        """
        return dict()

    def _define_dict_for_read(self):
        """Define a dictionary of functions to read HL properties from LL PVs.

        The dictionary defined in this function is called by read method.
        """
        return dict()

    def _get_from_pvs(self, is_sp, ll_prop, def_val=None):
        pv = self._writepvs[ll_prop] if is_sp else self._readpvs[ll_prop]
        if not pv.connected:
            return def_val
        val = pv.get(timeout=_CONN_TIMEOUT)
        return def_val if val is None else val

    def _start_lock_thread(self, pvname):
        # I have loop here to guarantee that the hardware will go to the
        # desired state.
        # I have to do this because of a problem on the LL IOCs triggered by
        # write commands which do not wait for the put operation to be
        # completed, which is the case for the default behavior of pyepics,
        # pydm, cs-studio...
        # This problem happens when one try to set different properties of the
        # LL IOCs, or the same property twice, in a time interval shorter than
        # the one it takes for LL IOC complete writing on the hardware (~60ms).
        maxatt = 100
        interv = 1 / 10  # I wait a little bit to reduce CPU load
        timer = self._lock_threads_dict.get(pvname)
        if timer is None or not timer.is_alive():
            timer = _Timer(
                interv, self._do_lock, args=(pvname, ), niter=maxatt)
            self._lock_threads_dict[pvname] = timer
            timer.start()
        else:
            timer.reset()

    def _do_lock(self, pvname):
        timer = self._lock_threads_dict[pvname]

        prop = self._dict_convert_pv2prop[pvname]
        my_val = self._config_ok_values.get(prop)
        pvo = self._writepvs.get(prop)
        if pvo is None:
            return
        if my_val is None:
            timer.stop()
        value = self._get_from_pvs(False, prop)
        if value is None:
            return
        if pvo.initialized and my_val == value:
            timer.stop()
        self._put_on_pv(pvo, my_val)

        if timer.cur_iter == timer.niters-1:
            _log.error(
                f'Could not set PV: {pvo.pvname:s} --> '
                f'my_val: {str(my_val):s} val: {str(value):s} '
                f'put_comp: {str(pvo.put_complete):s} '
                f'initia: {str(pvo.initialized):s} '
                f'conn: {str(pvo.connected):s}')

    def _put_on_pv(self, pvo, value, wait=False):
        if pvo.connected and pvo.put_complete is not False:
            # wait=True is too slow for the LL Timing IOCs. It is better not
            # to use it.
            try:
                pvo.put(value, use_complete=True, wait=wait)
                pvo.initialized = True
            except _CASeverityException:
                _log.error('NO Write Permission to {0:s}'.format(pvo.pvname))

    def _on_change_writepv(self, pvname, value, **kwargs):
        # -Cmd PVs do not have a state associated to them
        if value is None or _PVName.is_cmd_pv(pvname):
            return
        _Thread(target=self._on_change_pv_thread, args=(pvname, value)).start()

    def _on_change_readpv(self, pvname, value, **kwargs):
        if value is None:
            return
        if self._locked:
            self._start_lock_thread(pvname)
        _Thread(target=self._on_change_pv_thread, args=(pvname, value)).start()

        # at initialization load _config_ok_values
        prop = self._dict_convert_pv2prop.get(pvname)
        cond = prop is not None
        cond &= self._config_ok_values.get(prop) is None
        cond &= _PVName.is_rb_pv(pvname)
        cond &= not self._locked
        if cond:
            self._config_ok_values[prop] = value

    def _on_change_pv_thread(self, pvname, value, **kwargs):
        if not self._initialized:
            return

        pvn = _PVName.from_sp2rb(pvname)
        is_sp = _PVName.is_sp_pv(pvname)
        fun = self._dict_functs_for_update[self._dict_convert_pv2prop[pvn]]
        props = fun(is_sp, value)
        for hl_prop, val in props.items():
            if val is not None:
                self.run_callbacks(self.channel, hl_prop, val, is_sp=is_sp)

    def _on_connection_writepv(self, pvname, conn, **kwargs):
        if _PVName.is_cmd_pv(pvname):  # -Cmd must not change
            return
        prop = self._dict_convert_pv2prop[_PVName.from_sp2rb(pvname)]
        self._writepvs[prop].initialized = False  # not self._locked
        self._on_connection(pvname, conn)

    def _on_connection(self, pvname, conn, **kwargs):
        if not self._initialized:
            return
        self.run_callbacks(self.channel, None, None)

    def _set_simple(self, prop, value):
        """Simple setting of Low Level IOC PVs.

        Function called by write when no conversion is needed between
        high and low level properties.
        """
        if value is None:
            return dict()
        return {prop: value}

    def _get_simple(self, prop, is_sp, val=None, hl_prop=None):
        hl_prop = prop if hl_prop is None else hl_prop
        if val is None:
            val = self._get_from_pvs(is_sp, prop)
        if val is None:
            return dict()
        return {hl_prop: val}


class _EVROUT(_BaseLL):
    _REMOVE_PROPS = {}

    def __init__(self, channel, source_enums):
        fout_chan = _LLSearch.get_fout_channel(channel)
        self._fout_out = int(fout_chan.propty[3:])
        evg_chan = _LLSearch.get_evg_channel(channel)
        self._evg_out = int(evg_chan.propty[3:])
        self._source_enums = source_enums
        self._duration = None  # I keep this to avoid rounding errors

        prefix = LL_PREFIX + ('-' if LL_PREFIX else '')
        prefix += _PVName(channel).device_name + ':'
        super().__init__(channel, prefix)
        # self._config_ok_values['DevEnbl'] = 1
        # self._config_ok_values['FoutDevEnbl'] = 1
        # self._config_ok_values['EVGDevEnbl'] = 1
        if self.channel.propty.startswith('OUT'):
            intrg = _LLSearch.get_channel_internal_trigger_pvname(self.channel)
            intrg = int(intrg.propty[-2:])  # get internal trigger number
            self._config_ok_values['SrcTrig'] = intrg
            # Stop using FineDelay and RF Delay to ease consistency:
            self._config_ok_values['FineDelay'] = 0
            self._config_ok_values['RFDelay'] = 0

    def write(self, prop, value):
        # keep this info for recalculating Width whenever necessary
        if prop == 'Duration':
            self._duration = value
        return super().write(prop, value)

    def _define_convertion_prop2pv(self):
        intlb = _LLSearch.get_channel_internal_trigger_pvname(self.channel)
        outlb = _LLSearch.get_channel_output_port_pvname(self.channel)
        intlb = intlb.propty
        outlb = outlb.propty

        prefix = LL_PREFIX + ('-' if LL_PREFIX else '')
        evg_chan = _LLSearch.get_evg_channel(self.channel)
        _evg_prefix = prefix + evg_chan.device_name + ':'
        fout_chan = _LLSearch.get_fout_channel(self.channel)
        _fout_prefix = prefix + fout_chan.device_name + ':'
        map_ = {
            'State': self.prefix + intlb + 'State-Sts',
            'Evt': self.prefix + intlb + 'Evt-RB',
            'Width': self.prefix + intlb + 'WidthRaw-RB',
            'Polarity': self.prefix + intlb + 'Polarity-Sts',
            'NrPulses': self.prefix + intlb + 'NrPulses-RB',
            'Delay': self.prefix + intlb + 'DelayRaw-RB',
            'Src': self.prefix + outlb + 'Src-Sts',
            'SrcTrig': self.prefix + outlb + 'SrcTrig-RB',
            'RFDelay': self.prefix + outlb + 'RFDelayRaw-RB',
            'FineDelay': self.prefix + outlb + 'FineDelayRaw-RB',
            'RFDelayType': self.prefix + outlb + 'RFDelayType-Sts',
            # connection status PVs
            'DevEnbl': self.prefix + 'DevEnbl-Sts',
            'Network': self.prefix + 'Network-Mon',
            'Link': self.prefix + 'LinkStatus-Mon',
            'Intlk': self.prefix + 'IntlkStatus-Mon',
            'Los': self.prefix + 'Los-Mon',
            'EVGLos': _evg_prefix + 'Los-Mon',
            'FoutLos': _fout_prefix + 'Los-Mon',
            'FoutDevEnbl': _fout_prefix + 'DevEnbl-Sts',
            'EVGDevEnbl': _evg_prefix + 'DevEnbl-Sts',
            }
        for prop in self._REMOVE_PROPS:
            map_.pop(prop)
        return map_

    def _define_dict_for_write(self):
        map_ = {
            # 'DevEnbl': _partial(self._set_simple, 'DevEnbl'),
            # 'EVGDevEnbl': _partial(self._set_simple, 'EVGDevEnbl'),
            # 'FoutDevEnbl': _partial(self._set_simple, 'FoutDevEnbl'),
            'State': _partial(self._set_simple, 'State'),
            'Src': self._set_source,
            'Duration': self._set_duration,
            'Polarity': _partial(self._set_simple, 'Polarity'),
            'NrPulses': self._set_nrpulses,
            'Delay': _partial(self._set_delay, raw=False),
            'DelayRaw': _partial(self._set_delay, raw=True),
            'RFDelayType': _partial(self._set_simple, 'RFDelayType'),
            'RFDelayType': self._set_rfdelaytype,
            'LowLvlLock': self._set_locked,
            }
        return map_

    def _define_dict_for_update(self):
        map_ = {
            'State': _partial(self._get_simple, 'State'),
            'Evt': _partial(self._process_source, 'Evt'),
            'Width': _partial(self._get_duration_pulses, 'Width'),
            'Polarity': _partial(self._get_simple, 'Polarity'),
            'NrPulses': _partial(self._get_duration_pulses, 'NrPulses'),
            'Delay': _partial(self._get_delay, 'Delay'),
            'Src': _partial(self._process_source, 'Src'),
            'SrcTrig': _partial(self._process_source, 'SrcTrig'),
            'RFDelay': _partial(self._get_delay, 'RFDelay'),
            'FineDelay': _partial(self._get_delay, 'FineDelay'),
            'RFDelayType': _partial(self._get_simple, 'RFDelayType'),
            'DevEnbl': _partial(self._get_status, 'DevEnbl'),
            'Network': _partial(self._get_status, 'Network'),
            'Link': _partial(self._get_status, 'Link'),
            'Intlk': _partial(self._get_status, 'Intlk'),
            'Los': _partial(self._get_status, 'Los'),
            'EVGLos': _partial(self._get_status, 'EVGLos'),
            'FoutLos': _partial(self._get_status, 'FoutLos'),
            'EVGDevEnbl': _partial(self._get_status, 'EVGDevEnbl'),
            'FoutDevEnbl': _partial(self._get_status, 'FoutDevEnbl'),
            }
        for prop in self._REMOVE_PROPS:
            map_.pop(prop)
        return map_

    def _define_dict_for_read(self):
        map_ = {
            'State': _partial(self._get_simple, 'State'),
            'Duration': _partial(self._get_duration_pulses, ''),
            'Polarity': _partial(self._get_simple, 'Polarity'),
            'NrPulses': _partial(self._get_duration_pulses, ''),
            'Delay': _partial(self._get_delay, 'Delay'),
            'DelayRaw': _partial(self._get_delay, 'Delay'),
            'TotalDelay': _partial(self._get_delay, 'Delay'),
            'TotalDelayRaw': _partial(self._get_delay, 'Delay'),
            'Src': _partial(self._process_source, ''),
            'RFDelayType': _partial(self._get_simple, 'RFDelayType'),
            'Status': _partial(self._get_status, ''),
            'InInjTable': _partial(self._get_status, ''),
            'LowLvlLock': lambda is_sp: {'LowLvlLock': self.locked},
            }
        return map_

    def _get_status(self, prop, is_sp, value=None):
        dic_ = dict()
        dic_['DevEnbl'] = self._get_from_pvs(is_sp, 'DevEnbl', def_val=0)
        dic_['EVGDevEnbl'] = self._get_from_pvs(is_sp, 'EVGDevEnbl', def_val=0)
        dic_['FoutDevEnbl'] = self._get_from_pvs(
            is_sp, 'FoutDevEnbl', def_val=0)
        dic_['Network'] = self._get_from_pvs(False, 'Network', def_val=0)
        dic_['Link'] = self._get_from_pvs(False, 'Link', def_val=0)
        dic_['PVsConn'] = self.connected

        dic_['Intlk'] = 0
        if 'Intlk' not in self._REMOVE_PROPS:
            dic_['Intlk'] = self._get_from_pvs(False, 'Intlk', def_val=1)

        prt_num = 0
        dic_['Los'] = 0b00000000
        if 'Los' not in self._REMOVE_PROPS:
            prt_num = int(self.channel[-1])  # get OUT number for EVR
            dic_['Los'] = self._get_from_pvs(
                False, 'Los', def_val=0b11111111)
        dic_['EVGLos'] = self._get_from_pvs(
            False, 'EVGLos', def_val=0b11111111)
        dic_['FoutLos'] = self._get_from_pvs(
            False, 'FoutLos', def_val=0b11111111)

        if value is not None:
            dic_[prop] = value

        dic_['Los'] = _get_bit(dic_['Los'], prt_num)
        dic_['EVGLos'] = _get_bit(dic_['EVGLos'], self._evg_out)
        dic_['FoutLos'] = _get_bit(dic_['FoutLos'], self._fout_out)

        prob, bit = 0, 0
        prob, bit = _update_bit(prob, bit, not dic_['PVsConn']), bit+1
        prob, bit = _update_bit(prob, bit, not dic_['DevEnbl']), bit+1
        prob, bit = _update_bit(prob, bit, not dic_['FoutDevEnbl']), bit+1
        prob, bit = _update_bit(prob, bit, not dic_['EVGDevEnbl']), bit+1
        prob, bit = _update_bit(prob, bit, not dic_['Network']), bit+1
        prob, bit = _update_bit(prob, bit, not dic_['Link']), bit+1
        prob, bit = _update_bit(prob, bit, dic_['Los']), bit+1
        prob, bit = _update_bit(prob, bit, dic_['FoutLos']), bit+1
        prob, bit = _update_bit(prob, bit, dic_['EVGLos']), bit+1
        prob, bit = _update_bit(prob, bit, dic_['Intlk']), bit+1

        dic = {'Status': prob}
        dic.update(self._get_in_inj_table())
        return dic

    def _get_in_inj_table(self):
        src = self._process_source('', False).get('Src')
        if src is None:
            return dict()
        src_str = self._source_enums[src]
        ininj = False
        if src_str in self._events:
            ininj = self._events[src_str].is_in_inj_table
        ininj &= self._get_simple('State', False).get('State', False)
        return {'InInjTable': int(ininj)}

    def _get_delay(self, prop, is_sp, value=None):
        dic_ = dict()
        dic_['Delay'] = self._get_from_pvs(is_sp, 'Delay')
        dic_['RFDelay'] = self._get_from_pvs(is_sp, 'RFDelay', def_val=0)
        dic_['FineDelay'] = self._get_from_pvs(is_sp, 'FineDelay', def_val=0)
        dic_['RFDelayType'] = self._get_from_pvs(
            is_sp, 'RFDelayType', def_val=0)
        if value is not None:
            dic_[prop] = value
        if dic_['Delay'] is None:
            return dict()
        delay = dic_['Delay']*self.base_del + dic_['FineDelay']*_FDEL
        if not dic_['RFDelayType']:
            delay += dic_['RFDelay']*self._rf_del
        dic = {'Delay': delay, 'DelayRaw': dic_['Delay']}
        if not is_sp:
            dic = self._get_total_delay(dic)
        return dic

    def _get_total_delay(self, dic):
        src = self._process_source('', False).get('Src')
        if src is None:
            return dic
        src_str = self._source_enums[src]
        evt_del = 0
        if src_str in self._events:
            evt = self._events[src_str]
            evt_del = evt.delay_raw if evt.is_in_inj_table else 0
        evt_del = evt_del or 0  # in case event PV is disconnected
        dic['TotalDelayRaw'] = dic['DelayRaw'] + evt_del
        dic['TotalDelay'] = dic['Delay'] + evt_del*self.base_del
        return dic

    def _set_delay(self, value, raw=False):
        dic_ = {'RFDelay': 0, 'FineDelay': 0}
        if self._config_ok_values.get('RFDelayType', False):
            dic_['RFDelay'] = 31
        if value is None:
            return dic_
        value = value if raw else round(value / self.base_del)
        dic_['Delay'] = int(value)
        return dic_

    def _process_source(self, prop, is_sp, value=None):
        dic_ = {'SrcTrig': 30, 'Src': None, 'Evt': None}
        if 'SrcTrig' not in self._REMOVE_PROPS:
            dic_['SrcTrig'] = self._get_from_pvs(is_sp, 'SrcTrig')
        if 'Src' not in self._REMOVE_PROPS:
            dic_['Src'] = self._get_from_pvs(is_sp, 'Src')
        if 'Evt' not in self._REMOVE_PROPS:
            dic_['Evt'] = self._get_from_pvs(is_sp, 'Evt')
        if value is not None:
            dic_[prop] = value

        if any(map(lambda x: x is None, dic_.values())):
            return dict()

        ret = self._process_src_trig(dic_['SrcTrig'], is_sp)
        if ret is not None:
            return ret
        ret = self._process_src(dic_['Src'], is_sp)
        if ret is not None:
            return ret
        return self._process_evt(dic_['Evt'], is_sp)

    def _process_evt(self, evt, _):
        invalid = len(self._source_enums)-1  # Invalid option
        if evt not in _TIConst.EvtLL:
            return {'Src': invalid}
        evt_st = _TIConst.EvtLL._fields[_TIConst.EvtLL.index(evt)]
        if evt_st not in _TIConst.EvtLL2HLMap or \
                _TIConst.EvtLL2HLMap[evt_st] not in self._source_enums:
            return {'Src': invalid}
        else:
            ev_num = self._source_enums.index(_TIConst.EvtLL2HLMap[evt_st])
            return {'Src': ev_num}

    def _process_src_trig(self, src_trig, _):
        invalid = len(self._source_enums)-1  # Invalid option
        intrg = _LLSearch.get_channel_internal_trigger_pvname(self.channel)
        intrg = int(intrg.propty[-2:])  # get internal trigger number for EVR
        if src_trig != intrg:
            return {'Src': invalid}

    def _process_src(self, src, _):
        invalid = len(self._source_enums)-1  # Invalid option
        if src is None:
            return {'Src': invalid}

        # BUG: I noticed that differently from the EVR and EVE IOCs,
        # the AMCFPGAEVR do not have a 'Dsbl' as first option of the enums
        # list. So I have to create this offset to fix this...
        offset = 0
        if self.channel.dev.startswith('AMCFPGAEVR'):
            offset = 1
        try:
            source = _TIConst.TrigSrcLL._fields[src+offset]
        except IndexError:
            source = ''
        if not source:
            return {'Src': invalid}
        elif source.startswith(('Dsbl', 'Clock')):
            return {'Src': self._source_enums.index(source)}

    def _set_source(self, value):
        if value is None:
            return dict()
        # BUG: I noticed that differently from the EVR and EVE IOCs,
        # the AMCFPGAEVR do not have a 'Dsbl' as first option of the enums
        # list. So I have to create this offset to fix this...
        offset = int(self.channel.dev.startswith('AMCFPGAEVR'))

        if value >= (len(self._source_enums)-1):
            return dict()
        pname = self._source_enums[value]
        n = _TIConst.TrigSrcLL._fields.index('Trigger')
        if pname.startswith('Dsbl'):
            dic_ = {'Src': n, 'Evt': _TIConst.EvtLL.Evt00}
        if pname.startswith('Clock'):
            n = _TIConst.TrigSrcLL._fields.index(pname)
            n -= offset
            dic_ = {'Src': n}
        else:
            n -= offset
            evt = int(_TIConst.EvtHL2LLMap[pname][-2:])
            dic_ = {'Src': n, 'Evt': evt}
        if 'SrcTrig' in self._dict_convert_prop2pv.keys():
            intrg = _LLSearch.get_channel_internal_trigger_pvname(
                self.channel)
            intrg = int(intrg[-2:])  # get internal trigger number for EVR
            dic_['SrcTrig'] = intrg
        return dic_

    def _get_duration_pulses(self, prop, is_sp, value=None):
        dic_ = dict()
        dic_['NrPulses'] = self._get_from_pvs(is_sp, 'NrPulses')
        dic_['Width'] = self._get_from_pvs(is_sp, 'Width')
        for k, v in dic_.items():
            if v == 0:  # BUG: handle cases where LL sets these value to 0
                dic_[k] = 1
        if value is not None:
            dic_[prop] = value
        if any(map(lambda x: x is None, dic_.values())):
            return dict()
        return {
            'Duration': 2*dic_['Width']*self.base_del*dic_['NrPulses'],
            'NrPulses': dic_['NrPulses'],
            }

    def _set_duration(self, value, pul=None):
        if value is None:
            return dict()
        pul = pul or self._config_ok_values.get('NrPulses')
        pul = pul or 1  # BUG: handle cases where LL sets this value to 0
        wid = value / self.base_del / pul / 2
        wid = round(wid) if wid >= 1 else 1
        return {'Width': wid}

    def _set_nrpulses(self, pul):
        if pul is None or pul < 1:
            return dict()
        pul = int(pul)
        dic = {'NrPulses': pul}

        # at initialization, try to set _duration
        if self._duration is None:
            # BUG: handle cases where LL sets these value to 0
            wid = self._config_ok_values.get('Width') or 1
            self._duration = wid * pul * 2 * self.base_del

        if self._duration is not None:
            dic.update(self._set_duration(self._duration, pul=pul))
        return dic

    def _set_rfdelaytype(self, value):
        """Simple setting of Low Level IOC PVs.

        Function called by write when no conversion is needed between
        high and low level properties.
        """
        if value is None:
            return dict()
        dic = {'RFDelayType': value, 'RFDelay': 0}
        if value:
            dic['RFDelay'] = 31
        return dic


class _EVROTP(_EVROUT):
    _REMOVE_PROPS = {
        'RFDelay', 'FineDelay', 'Src', 'SrcTrig', 'RFDelayType', 'Los'}

    def _get_delay(self, prop, is_sp, val=None):
        if val is None:
            val = self._get_from_pvs(is_sp, 'Delay')
        if val is None:
            return dict()
        dic = {'Delay': val * self.base_del, 'DelayRaw': val}
        if not is_sp:
            dic = self._get_total_delay(dic)
        return dic

    def _set_delay(self, value, raw=False):
        return {'Delay': int(value if raw else round(value / self.base_del))}

    def _process_source(self, prop, is_sp, val=None):
        if val is None:
            val = self._get_from_pvs(is_sp, 'Evt')
        if val is None:
            return dict()
        return self._process_evt(val, is_sp)

    def _process_src(self, src, _):
        return dict()

    def _set_source(self, value):
        if value >= (len(self._source_enums)-1):
            return dict()
        pname = self._source_enums[value]
        dic_ = dict()
        reg = _re.compile('[0-9]+')
        if pname.startswith('Dsbl'):
            dic_['Evt'] = 0
        if not pname.startswith('Clock'):
            mat = reg.findall(_TIConst.EvtHL2LLMap[pname])
            dic_['Evt'] = int(mat[0])
        return dic_


class _EVEOTP(_EVROTP):
    pass


class _EVEOUT(_EVROUT):
    _REMOVE_PROPS = {'Los', }


class _AMCFPGAEVRAMC(_EVROUT):
    _REMOVE_PROPS = {
        'RFDelay', 'FineDelay', 'SrcTrig', 'RFDelayType', 'Intlk', 'Los'}

    def _get_delay(self, prop, is_sp, value=None):
        return _EVROTP._get_delay(self, prop, is_sp, value)

    def _set_delay(self, value, raw=False):
        return _EVROTP._set_delay(self, value, raw=raw)

    def _define_convertion_prop2pv(self):
        map_ = super()._define_convertion_prop2pv()
        map_['Network'] = map_['Network'].replace('Network', 'RefClkLocked')
        return map_

    def _process_source(self, prop, is_sp, value=None):
        dic_ = dict()
        dic_['Src'] = self._get_from_pvs(is_sp, 'Src')
        dic_['Evt'] = self._get_from_pvs(is_sp, 'Evt')
        if value is not None:
            dic_[prop] = value
        if any(map(lambda x: x is None, dic_.values())):
            return dict()
        ret = self._process_src(dic_['Src'], is_sp)
        if ret is not None:
            return ret
        return self._process_evt(dic_['Evt'], is_sp)


class _AMCFPGAEVRFMC(_AMCFPGAEVRAMC):
    pass


def get_ll_trigger(channel, source_enums):
    """Get Low Level trigger objects."""
    LL_TRIGGER_CLASSES = {
        ('EVR', 'OUT'): _EVROUT,
        ('EVR', 'OTP'): _EVROTP,
        ('EVE', 'OTP'): _EVEOTP,
        ('EVE', 'OUT'): _EVEOUT,
        ('AMCFPGAEVR', 'CRT'): _AMCFPGAEVRAMC,
        ('AMCFPGAEVR', 'FMC'): _AMCFPGAEVRFMC,
        }
    chan = _PVName(channel)
    key = (chan.dev, chan.propty[:3])
    cls_ = LL_TRIGGER_CLASSES.get(key)
    if not cls_:
        raise Exception(
            'Low Level Trigger Class not defined for device ' +
            'type '+key[0]+' and connection type '+key[1]+'.')
    return cls_(channel, source_enums)



"""."""

from .device import Device as _Device


class BeamlineScreen(_Device):
    """."""

    class DEVICES:
        """Devices names."""

        MANACA = 'SI-09SABL:AP-Manaca-MVS2:'
        ALL = (MANACA, )

    _properties = (
        'BeamCenterX-Mon', 'BeamCentermmX-Mon',
        'BeamCenterY-Mon', 'BeamCentermmY-Mon',
        'BeamSizeX-Mon', 'BeamSizemmX-Mon',
        'BeamSizeY-Mon', 'BeamSizemmY-Mon',
    )

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in BeamlineScreen.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=BeamlineScreen._properties)

    @property
    def centerx_pxl(self):
        """."""
        return self['BeamCenterX-Mon']

    @property
    def centerx_mm(self):
        """."""
        return self['BeamCentermmX-Mon']

    @property
    def centery_pxl(self):
        """."""
        return self['BeamCenterY-Mon']

    @property
    def centery_mm(self):
        """."""
        return self['BeamCentermmY-Mon']

    @property
    def sizex_pxl(self):
        """."""
        return self['BeamSizeX-Mon']

    @property
    def sizex_mm(self):
        """."""
        return self['BeamSizemmX-Mon']

    @property
    def sizey_pxl(self):
        """."""
        return self['BeamSizeY-Mon']

    @property
    def sizey_mm(self):
        """."""
        return self['BeamSizemmY-Mon']



#!/usr/local/bin/python-sirius
"""PS Diag PVs."""

import numpy as _np

from ...util import get_bit as _get_bit
from ...namesys import SiriusPVName as _PVName
from ...search import PSSearch as _PSSearch
from ...pwrsupply.csdev import Const as _PSConst, ETypes as _ETypes, \
    PS_LI_INTLK_THRS as _PS_LI_INTLK_THRS


class PSDiffPV:
    """Diff of a PS current setpoint and a readback."""

    CURRT_SP = 0
    CURRT_MON = 1

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute difference between SP and Mon current values."""
        disconnected = \
            not computed_pv.pvs[PSDiffPV.CURRT_SP].connected or \
            not computed_pv.pvs[PSDiffPV.CURRT_MON].connected
        if disconnected:
            return None
        value_sp = computed_pv.pvs[PSDiffPV.CURRT_SP].value
        value_mon = computed_pv.pvs[PSDiffPV.CURRT_MON].value
        diff = value_mon - value_sp
        return {'value': diff}


class PSStatusPV:
    """Power Supply Status PV."""

    BIT_PSCONNECT = 0b0000001
    BIT_PWRSTATON = 0b0000010
    BIT_CURRTDIFF = 0b0000100
    BIT_INTERLOCK = 0b0001000
    BIT_ALARMSSET = 0b0010000
    BIT_OPMODEDIF = 0b0100000
    BIT_BOWFMDIFF = 0b1000000

    PWRSTE_STS = 0
    CURRT_DIFF = 1
    INTRLCK_LI = 2
    WARNSTS_LI = 3
    CONNCTD_LI = 4
    OPMODE_SEL = 2
    OPMODE_STS = 3
    WAVFRM_MON = 4

    DTOLWFM_DICT = dict()

    def __init__(self):
        """Init attributs."""
        self.INTLK_PVS = list()
        self.ALARM_PVS = list()
        self.intlkwarn_bit = _ETypes.LINAC_INTLCK_WARN.index('LoadI Over Thrs')

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute PS Status PV."""
        psname = _PVName(computed_pv.pvs[0].pvname).device_name
        value = 0
        # ps connected?
        if psname.sec != 'LI':
            disconn = \
                not computed_pv.pvs[PSStatusPV.PWRSTE_STS].connected or \
                not computed_pv.pvs[PSStatusPV.CURRT_DIFF].connected or \
                not computed_pv.pvs[PSStatusPV.OPMODE_SEL].connected or \
                not computed_pv.pvs[PSStatusPV.OPMODE_STS].connected
            if psname.dev not in ['FCH', 'FCV']:
                disconn |= not computed_pv.pvs[PSStatusPV.WAVFRM_MON].connected

            for intlk in self.INTLK_PVS:
                disconn |= not computed_pv.pvs[intlk].connected
            for alarm in self.ALARM_PVS:
                disconn |= not computed_pv.pvs[alarm].connected

            if not disconn and psname.dev not in ['FCH', 'FCV']:  # comm ok?
                commsts = computed_pv.pvs[PSStatusPV.PWRSTE_STS].status
                if commsts != 0 or commsts is None:
                    disconn = True
        else:
            disconn = \
                not computed_pv.pvs[PSStatusPV.PWRSTE_STS].connected or \
                not computed_pv.pvs[PSStatusPV.CURRT_DIFF].connected or \
                not computed_pv.pvs[PSStatusPV.INTRLCK_LI].connected

            if not disconn:  # comm ok?
                commval = computed_pv.pvs[PSStatusPV.CONNCTD_LI].value
                commsts = computed_pv.pvs[PSStatusPV.CONNCTD_LI].status
                if commval != 0 or commval is None or commsts != 0:
                    disconn = True
        if disconn:
            value |= PSStatusPV.BIT_PSCONNECT
            value |= PSStatusPV.BIT_PWRSTATON
            value |= PSStatusPV.BIT_CURRTDIFF
            value |= PSStatusPV.BIT_INTERLOCK
            value |= PSStatusPV.BIT_ALARMSSET
            value |= PSStatusPV.BIT_OPMODEDIF
            value |= PSStatusPV.BIT_BOWFMDIFF
            return {'value': value}

        # pwrstate?
        pwrsts = computed_pv.pvs[PSStatusPV.PWRSTE_STS].value
        if pwrsts != _PSConst.PwrStateSts.On or pwrsts is None:
            value |= PSStatusPV.BIT_PWRSTATON

        if psname.sec != 'LI':
            # opmode?
            sel = computed_pv.pvs[PSStatusPV.OPMODE_SEL].value
            sts = computed_pv.pvs[PSStatusPV.OPMODE_STS].value
            if sel is not None and sts is not None:
                if psname.dev in ['FCH', 'FCV']:
                    opmode_sel = _ETypes.FOFB_OPMODES[sel]
                    opmode_sts = _ETypes.FOFB_OPMODES[sts]
                    checkdiff = sts == _PSConst.OpModeFOFB.closed_loop_manual
                else:
                    opmode_sel = _ETypes.OPMODES[sel]
                    opmode_sts = _ETypes.STATES[sts]
                    checkdiff = sts == _PSConst.States.SlowRef
                if opmode_sel != opmode_sts:
                    value |= PSStatusPV.BIT_OPMODEDIF
                # current-diff?
                if checkdiff:
                    severity = computed_pv.pvs[PSStatusPV.CURRT_DIFF].severity
                    if severity != 0:
                        value |= PSStatusPV.BIT_CURRTDIFF
                # waveform diff?
                elif (psname.sec == 'BO') and (sts == _PSConst.States.RmpWfm):
                    mon = computed_pv.pvs[PSStatusPV.WAVFRM_MON].value
                    if psname not in PSStatusPV.DTOLWFM_DICT.keys():
                        pstype = _PSSearch.conv_psname_2_pstype(psname)
                        PSStatusPV.DTOLWFM_DICT[psname] = _PSSearch.get_splims(
                            pstype, 'DTOL_WFM')
                    if not _np.allclose(
                            mon, 0.0, atol=PSStatusPV.DTOLWFM_DICT[psname]):
                        value |= PSStatusPV.BIT_BOWFMDIFF
            else:
                value |= PSStatusPV.BIT_OPMODEDIF

            # interlocks?
            for intlk in self.INTLK_PVS:
                intlkval = computed_pv.pvs[intlk].value
                if intlkval != 0 or intlkval is None:
                    value |= PSStatusPV.BIT_INTERLOCK
                    break

            # alarms?
            for alarm in self.ALARM_PVS:
                alarmval = computed_pv.pvs[alarm].value
                if alarmval != 0 or alarmval is None:
                    value |= PSStatusPV.BIT_ALARMSSET
                    break

        else:
            # current-diff?
            severity = computed_pv.pvs[PSStatusPV.CURRT_DIFF].severity
            if severity != 0:
                value |= PSStatusPV.BIT_CURRTDIFF

            # interlocks?
            intlk = computed_pv.pvs[PSStatusPV.INTRLCK_LI].value
            if psname.dev == 'Spect':
                intlkwarn = computed_pv.pvs[PSStatusPV.WARNSTS_LI].value
                if _get_bit(intlkwarn, self.intlkwarn_bit):
                    intlk -= 2**self.intlkwarn_bit
            if intlk > _PS_LI_INTLK_THRS or intlk is None:
                value |= PSStatusPV.BIT_INTERLOCK

        return {'value': value}



"""Define PVs, constants and properties of SI AP Manaca IOC."""
from .. import csdev as _csdev


# --- Enumeration Types ---
class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    MEASURESTATE = ('Stopped', 'Measuring')
    APPLY_STATUS = ('Idle', 'Applied', 'ConnectionError', 'LimitExceeded')


_et = ETypes  # syntactic sugar


# --- Const class ---
class Const(_csdev.Const):
    """Const class defining constants."""

    DEF_COEFX = 10.00 * 1e-3  # [mm/px] (source: Sergio Lordano from OPT)
    DEF_COEFY = 10.04 * 1e-3  # [mm/px] (source: Sergio Lordano from OPT)
    DEF_ROISIZE = 150
    TARGETX = 375  # [px] source: Lucas Sanfelici via email (was 372)
    TARGETY = 720  # [px] source: Lucas Sanfelici via email
    MAX_BUMP = 10  # [urad]
    DIST_FROM_SRC = 30.160  # [m] source: Sergio Lordano from OPT
    DEF_PROFILE = 'MNC:A:BASLER02:'
    DEF_RATE = 5  # [Hz]
    IP_IOC = '10.31.74.45'
    PREFIX_IOC = 'SI-09SABL:AP-Manaca-MVS2:'
    SS_INDEX = 9
    SS_SIZE = 7.0358  # [m]
    BPM1_INDEX = (SS_INDEX-1) * 8 - 1  # 63
    BPM2_INDEX = (SS_INDEX-1) * 8  # 64

    MeasureState = _csdev.Const.register('MeasureState', _et.MEASURESTATE)
    ApplyStatus = _csdev.Const.register('ApplyStatus', _et.APPLY_STATUS)

    @classmethod
    def get_database(cls, prefix=''):
        """Return IOC database."""
        dbase = {
            'ApplyBump-Cmd': {'type': 'int', 'value': 0},
            'MeasureCtrl-Sel': {
                'type': 'enum', 'value': cls.MeasureState.Stopped,
                'enums': cls.MeasureState._fields},
            'MeasureCtrl-Sts': {
                'type': 'enum', 'value': cls.MeasureState.Stopped,
                'enums': cls.MeasureState._fields},
            'MeasureRate-SP': {
                'type': 'float', 'prec': 4, 'unit': 'Hz',
                'value': cls.DEF_RATE, 'lolim': 0.1, 'hilim': 30},
            'MeasureRate-RB': {
                'type': 'float', 'prec': 4, 'unit': 'Hz',
                'value': cls.DEF_RATE, 'lolim': 0.1, 'hilim': 30},
            'TargetPosX-SP': {
                'type': 'int', 'value': cls.TARGETX, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'TargetPosX-RB': {
                'type': 'int', 'value': cls.TARGETX, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'TargetPosY-SP': {
                'type': 'int', 'value': cls.TARGETY, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'TargetPosY-RB': {
                'type': 'int', 'value': cls.TARGETY, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'NeededDeltaBumpX-Mon': {
                'type': 'float', 'prec': 3, 'unit': 'urad',
                'value': 0.0, 'lolim': -10, 'hilim': 10},
            'NeededDeltaBumpY-Mon': {
                'type': 'float', 'prec': 3, 'unit': 'urad',
                'value': 0.0, 'lolim': -10, 'hilim': 10},
            'AppliedBumpX-Mon': {
                'type': 'float', 'prec': 3, 'unit': 'urad',
                'value': 0.0, 'lolim': -10, 'hilim': 10},
            'AppliedBumpY-Mon': {
                'type': 'float', 'prec': 3, 'unit': 'urad',
                'value': 0.0, 'lolim': -10, 'hilim': 10},
            'ApplyStatus-Mon': {
                'type': 'enum', 'value': cls.ApplyStatus.Idle,
                'enums': cls.ApplyStatus._fields,
                'states': (0, 0, 2, 2)},
            }
        for val in dbase.values():
            low = val.get('lolim', None)
            hig = val.get('hilim', None)
            if low is not None:
                val['lolo'] = low
                val['low'] = low
            if hig is not None:
                val['hihi'] = hig
                val['high'] = hig
        if prefix:
            dbase = {prefix + k: v for k, v in dbase.items()}
        dbase.update(_csdev.Const.get_database(prefix=prefix))

        dbase[prefix+'ImgFlipY-Sel']['value'] = cls.ImgFlip.Off
        dbase[prefix+'ImgFlipY-Sts']['value'] = cls.ImgFlip.Off
        dbase[prefix+'ImgFlipX-Sel']['value'] = cls.ImgFlip.On
        dbase[prefix+'ImgFlipX-Sts']['value'] = cls.ImgFlip.On
        dbase[prefix+'ROISizeX-SP']['value'] = cls.DEF_ROISIZE
        dbase[prefix+'ROISizeX-RB']['value'] = cls.DEF_ROISIZE
        dbase[prefix+'ROISizeY-SP']['value'] = cls.DEF_ROISIZE
        dbase[prefix+'ROISizeY-RB']['value'] = cls.DEF_ROISIZE
        dbase[prefix+'ROICenterX-SP']['value'] = cls.TARGETX
        dbase[prefix+'ROICenterX-RB']['value'] = cls.TARGETX
        dbase[prefix+'ROICenterY-SP']['value'] = cls.TARGETY
        dbase[prefix+'ROICenterY-RB']['value'] = cls.TARGETY
        dbase[prefix+'ROIAutoCenter-Sel']['value'] = cls.AutoCenter.Manual
        dbase[prefix+'ROIAutoCenter-Sts']['value'] = cls.AutoCenter.Manual
        dbase[prefix+'CalcMethod-Sel']['value'] = cls.Method.Moments
        dbase[prefix+'CalcMethod-Sts']['value'] = cls.Method.Moments

        dbase.pop(prefix + 'Px2mmScaleX-SP')
        info = dbase.pop(prefix + 'Px2mmScaleX-RB')
        info['value'] = cls.DEF_COEFX
        dbase[prefix + 'Px2mmScaleX-Cte'] = info

        dbase.pop(prefix + 'Px2mmScaleY-SP')
        info = dbase.pop(prefix + 'Px2mmScaleY-RB')
        info['value'] = cls.DEF_COEFY
        dbase[prefix + 'Px2mmScaleY-Cte'] = info
        return dbase



"""Define PVs, constants and properties of Meas IOCs."""
import numpy as _np

from .. import csdev as _csdev


# --- Enumeration Types ---
class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    READINGORDER = ('CLike', 'FortranLike')
    METHOD = ('GaussFit', 'Moments')
    AUTO_CENTER = ('Manual', 'Auto')
    BG_CTRL = ('Acquire', 'Reset')
    BG_USAGE = ('NotUsing', 'Using')
    CROPIDX = ('Low', 'High')
    FLIP = ('Off', 'On')
    FITPARAMS = ('Amp', 'Cen', 'Sig', 'Off')
    PLANES = ('Y', 'X')


_et = ETypes  # syntactic sugar


# --- Const class ---
class Const(_csdev.Const):
    """Const class defining constants."""

    DEFAULT_WIDTH = 1024
    DEFAULT_HEIGHT = 1090
    DEFAULT_ROI_SIZE = 500
    MAX_WIDTH = 3000

    ReadingOrder = _csdev.Const.register('ReadingOrder', _et.READINGORDER)
    Method = _csdev.Const.register('Method', _et.METHOD)
    AutoCenter = _csdev.Const.register('AutoCenter', _et.AUTO_CENTER)
    BgCtrl = _csdev.Const.register('BgCtrl', _et.BG_CTRL)
    BgUsage = _csdev.Const.register('BgUsage', _et.BG_USAGE)
    CropIdx = _csdev.Const.register('CropIdx', _et.CROPIDX)
    ImgFlip = _csdev.Const.register('ImgFlip', _et.FLIP)
    FitParams = _csdev.Const.register('FitParams', _et.FITPARAMS)
    Plane = _csdev.Const.register('Plane', _et.PLANES)

    @classmethod
    def get_database(cls, prefix=''):
        """Return IOC database."""
        dbase = {
            'Log-Mon': {'type': 'char', 'value': '', 'count': 200},
            'Image-SP': {
                'type': 'int',
                'value': _np.zeros(
                    cls.DEFAULT_WIDTH*cls.DEFAULT_WIDTH, dtype=int),
                'count': cls.MAX_WIDTH*cls.MAX_WIDTH},
            'Image-RB': {
                'type': 'int',
                'value': _np.zeros(
                    cls.DEFAULT_WIDTH*cls.DEFAULT_WIDTH, dtype=int),
                'count': cls.MAX_WIDTH*cls.MAX_WIDTH},
            'Width-SP': {
                'type': 'int', 'value': cls.DEFAULT_WIDTH, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'Width-RB': {
                'type': 'int', 'value': cls.DEFAULT_WIDTH, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ReadingOrder-Sel': {
                'type': 'enum', 'value': cls.ReadingOrder.CLike,
                'enums': cls.ReadingOrder._fields},
            'ReadingOrder-Sts': {
                'type': 'enum', 'value': cls.ReadingOrder.CLike,
                'enums': cls.ReadingOrder._fields},
            'NrAverages-SP': {
                'type': 'int', 'unit': '#', 'value': 1,
                'lolim': 1, 'hilim': 999},
            'NrAverages-RB': {
                'type': 'int', 'unit': '#', 'value': 1,
                'lolim': 1, 'hilim': 999},
            'ResetBuffer-Cmd': {'type': 'int', 'value': 0},
            'BufferSize-Mon': {'type': 'int', 'value': 0},
            'ImgCropLow-SP': {
                'type': 'int', 'value': 0, 'unit': '',
                'lolim': 0, 'hilim': 2**16},
            'ImgCropLow-RB': {
                'type': 'int', 'value': 0, 'unit': '',
                'lolim': 0, 'hilim': 2**16},
            'ImgCropHigh-SP': {
                'type': 'int', 'value': 2**16, 'unit': '',
                'lolim': 0, 'hilim': 2**16},
            'ImgCropHigh-RB': {
                'type': 'int', 'value': 2**16, 'unit': '',
                'lolim': 0, 'hilim': 2**16},
            'ImgCropUse-Sel': {
                'type': 'enum', 'value': cls.BgUsage.NotUsing,
                'enums': cls.BgUsage._fields},
            'ImgCropUse-Sts': {
                'type': 'enum', 'value': cls.BgUsage.NotUsing,
                'enums': cls.BgUsage._fields},
            'ImgFlipX-Sel': {
                'type': 'enum', 'value': cls.ImgFlip.Off,
                'enums': cls.ImgFlip._fields},
            'ImgFlipX-Sts': {
                'type': 'enum', 'value': cls.ImgFlip.Off,
                'enums': cls.ImgFlip._fields},
            'ImgFlipY-Sel': {
                'type': 'enum', 'value': cls.ImgFlip.Off,
                'enums': cls.ImgFlip._fields},
            'ImgFlipY-Sts': {
                'type': 'enum', 'value': cls.ImgFlip.Off,
                'enums': cls.ImgFlip._fields},
            'CalcMethod-Sel': {
                'type': 'enum', 'value': cls.Method.GaussFit,
                'enums': cls.Method._fields},
            'CalcMethod-Sts': {
                'type': 'enum', 'value': cls.Method.GaussFit,
                'enums': cls.Method._fields},
            'ROIAutoCenter-Sel': {
                'type': 'enum', 'value': cls.AutoCenter.Auto,
                'enums': cls.AutoCenter._fields},
            'ROIAutoCenter-Sts': {
                'type': 'enum', 'value': cls.AutoCenter.Auto,
                'enums': cls.AutoCenter._fields},
            'ROICenterX-SP': {
                'type': 'int', 'value': 0, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROICenterX-RB': {
                'type': 'int', 'value': 0, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROICenterY-SP': {
                'type': 'int', 'value': 0, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROICenterY-RB': {
                'type': 'int', 'value': 0, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROISizeX-SP': {
                'type': 'int', 'value': cls.DEFAULT_ROI_SIZE, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROISizeX-RB': {
                'type': 'int', 'value': cls.DEFAULT_ROI_SIZE, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROISizeY-SP': {
                'type': 'int', 'value': cls.DEFAULT_ROI_SIZE, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROISizeY-RB': {
                'type': 'int', 'value': cls.DEFAULT_ROI_SIZE, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            # 'Background-SP': {
            #     'type': 'int',
            #     'value': _np.zeros(cls.DEFAULT_WIDTH*cls.DEFAULT_WIDTH),
            #     'count': cls.MAX_WIDTH*cls.MAX_WIDTH},
            # 'Background-RB': {
            #     'type': 'int',
            #     'value': _np.zeros(cls.DEFAULT_WIDTH*cls.DEFAULT_WIDTH),
            #     'count': cls.MAX_WIDTH*cls.MAX_WIDTH},
            # 'BgUse-Sel': {
            #     'type': 'enum', 'value': cls.BgUsage.NotUsing,
            #     'enums': cls.BgUsage._fields},
            # 'BgUse-Sts': {
            #     'type': 'enum', 'value': cls.BgUsage.NotUsing,
            #     'enums': cls.BgUsage._fields},
            'Px2mmScaleX-SP': {
                'type': 'float', 'value': 1, 'prec': 3,
                'unit': 'mm/px', 'lolim': 0, 'hilim': 10},
            'Px2mmScaleX-RB': {
                'type': 'float', 'value': 1, 'prec': 3,
                'unit': 'mm/px', 'lolim': 0, 'hilim': 10},
            'Px2mmScaleY-SP': {
                'type': 'float', 'value': 1, 'prec': 3,
                'unit': 'mm/px', 'lolim': 0, 'hilim': 10},
            'Px2mmScaleY-RB': {
                'type': 'float', 'value': 1, 'prec': 3,
                'unit': 'mm/px', 'lolim': 0, 'hilim': 10},
            'Px2mmAutoCenter-Sel': {
                'type': 'enum', 'value': cls.AutoCenter.Auto,
                'enums': cls.AutoCenter._fields},
            'Px2mmAutoCenter-Sts': {
                'type': 'enum', 'value': cls.AutoCenter.Auto,
                'enums': cls.AutoCenter._fields},
            'Px2mmCenterX-SP': {
                'type': 'int', 'value': 0,
                'unit': 'px', 'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'Px2mmCenterX-RB': {
                'type': 'int', 'value': 0,
                'unit': 'px', 'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'Px2mmCenterY-SP': {
                'type': 'int', 'value': 0,
                'unit': 'px', 'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'Px2mmCenterY-RB': {
                'type': 'int', 'value': 0,
                'unit': 'px', 'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROIStartX-Mon': {
                'type': 'int', 'value': 0, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROIStartY-Mon': {
                'type': 'int', 'value': 0, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROIEndX-Mon': {
                'type': 'int', 'value': cls.DEFAULT_ROI_SIZE, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROIEndY-Mon': {
                'type': 'int', 'value': cls.DEFAULT_ROI_SIZE, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'ROIProjX-Mon': {
                'type': 'int', 'value': _np.zeros(cls.DEFAULT_ROI_SIZE),
                'count': cls.MAX_WIDTH},
            'ROIProjY-Mon': {
                'type': 'int', 'value': _np.zeros(cls.DEFAULT_ROI_SIZE),
                'count': cls.MAX_WIDTH},
            'ROIAxisX-Mon': {
                'type': 'int', 'value': _np.arange(cls.DEFAULT_ROI_SIZE),
                'count': cls.MAX_WIDTH},
            'ROIAxisY-Mon': {
                'type': 'int', 'value': _np.arange(cls.DEFAULT_ROI_SIZE),
                'count': cls.MAX_WIDTH},
            'ROIGaussFitX-Mon': {
                'type': 'float', 'value': 0.0*_np.zeros(cls.DEFAULT_ROI_SIZE),
                'count': cls.MAX_WIDTH},
            'ROIGaussFitY-Mon': {
                'type': 'float', 'value': 0.0*_np.zeros(cls.DEFAULT_ROI_SIZE),
                'count': cls.MAX_WIDTH},
            'BeamCenterX-Mon': {
                'type': 'int', 'value': 1, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'BeamCenterY-Mon': {
                'type': 'int', 'value': 1, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'BeamSizeX-Mon': {
                'type': 'int', 'value': 1, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'BeamSizeY-Mon': {
                'type': 'int', 'value': 1, 'unit': 'px',
                'lolim': 0, 'hilim': cls.MAX_WIDTH},
            'BeamCentermmX-Mon': {
                'type': 'float', 'value': 1, 'prec': 3,
                'unit': 'mm', 'lolim': -20, 'hilim': 20},
            'BeamCentermmY-Mon': {
                'type': 'float', 'value': 1, 'prec': 3,
                'unit': 'mm', 'lolim': -20, 'hilim': 20},
            'BeamSizemmX-Mon': {
                'type': 'float', 'value': 1, 'prec': 3,
                'unit': 'mm', 'lolim': -20, 'hilim': 20},
            'BeamSizemmY-Mon': {
                'type': 'float', 'value': 1, 'prec': 3,
                'unit': 'mm', 'lolim': -20, 'hilim': 20},
            'BeamAmplX-Mon': {
                'type': 'float', 'value': 0, 'prec': 3,
                'unit': 'mm', 'lolim': -20, 'hilim': 20},
            'BeamAmplY-Mon': {
                'type': 'float', 'value': 0, 'prec': 3,
                'unit': 'mm', 'lolim': -20, 'hilim': 20},
            'BeamOffsetX-Mon': {
                'type': 'float', 'value': 0, 'prec': 3,
                'unit': 'mm', 'lolim': -20, 'hilim': 20},
            'BeamOffsetY-Mon': {
                'type': 'float', 'value': 0, 'prec': 3,
                'unit': 'mm', 'lolim': -20, 'hilim': 20},
            }
        for val in dbase.values():
            low = val.get('lolim', None)
            hig = val.get('hilim', None)
            if low is not None:
                val['lolo'] = low
                val['low'] = low
            if hig is not None:
                val['hihi'] = hig
                val['high'] = hig
        if prefix:
            return {prefix + k: v for k, v in dbase.items()}
        return dbase



"""BSMP Constants."""

# --- Reserved Address IDs ---
ID_ADDR_MASTER: int = 0
ID_ADDR_BROADCAST: int = 255

# --- Reserved Group IDs ---
ID_STD_GROUP_ALL: int = 0
ID_STD_GROUP_READONLY: int = 1
ID_STD_GROUP_WRITE: int = 2

# --- Query Commands ---
CMD_CMD_QUERY_PROTOCOL_VERSION: int = 0x00
CMD_PROTOCOL_VERSION: int = 0x01
CMD_CMD_QUERY_LIST_OF_VARIABLES: int = 0x02
CMD_LIST_OF_VARIABLE: int = 0x03
CMD_QUERY_LIST_OF_GROUP_OF_VARIABLES: int = 0x04
CMD_LIST_OF_GROUP_OF_VARIABLES: int = 0x05
CMD_QUERY_GROUP_OF_VARIABLES: int = 0x06
CMD_GROUP_OF_VARIABLES: int = 0x07
CMD_QUERY_LIST_OF_CURVES: int = 0x08
CMD_LIST_OF_CURVES: int = 0x09
CMD_QUERY_CURVE_CHECKSUM: int = 0x0A
CMD_CURVE_CHECKSUM: int = 0x0B
CMD_QUERY_LIST_OF_FUNCTIONS: int = 0x0C
CMD_LIST_OF_FUNCTIONS: int = 0x0D

# --- Reading Commands ---
CMD_READ_VARIABLE: int = 0x10
CMD_VARIABLE_VALUE: int = 0x11
CMD_READ_GROUP_OF_VARIABLES: int = 0x12
CMD_GROUP_OF_VARIABLES_VALUE: int = 0x13

# --- Writing Commands ---
CMD_WRITE_VARIABLE: int = 0x20
CMD_WRITE_GROUP_OF_VARIABLES: int = 0x22
CMD_BINARY_OPERATION_IN_A_VARIABLE: int = 0x24
CMD_BINARY_OPERATION_IN_A_GROUP: int = 0x26
CMD_WRITE_AND_READ_VARIABLES: int = 0x28

# --- Group of Variables Manipulation Commnads ---
CMD_CREATE_GROUP_OF_VARIABLES: int = 0x30
CMD_REMOVE_ALL_GROUPS_OF_VARIABLES: int = 0x32

# --- Curve Transfer Commands ---
CMD_REQUEST_CURVE_BLOCK: int = 0x40
CMD_CURVE_BLOCK: int = 0x41
CMD_RECALCULATE_CURVE_CHECKSUM: int = 0x42

# --- Function Execution Commands ---
CMD_EXECUTE_FUNCTION: int = 0x50
CMD_FUNCTION_RETURN: int = 0x51
CMD_FUNCTION_ERROR: int = 0x53

# --- Error Commands ---
ACK_OK: int = 0xE0
ACK_MALFORMED_MESSAGE: int = 0xE1
ACK_OPERATION_NOT_SUPPORTED: int = 0xE2
ACK_INVALID_ID: int = 0xE3
ACK_INVALID_VALUE: int = 0xE4
ACK_INVALID_PAYLOAD_SIZE: int = 0xE5
ACK_READ_ONLY: int = 0xE6
ACK_INSUFFICIENT_MEMORY: int = 0xE7
ACK_RESOURCE_BUSY: int = 0xE8
